#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) 2017, sookido
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = '''
---
module: zabbix_template
short_description: Create/update/delete/dump Zabbix template
description:
    - This module allows you to create, modify, delete and dump Zabbix templates.
    - Multiple templates can be created or modified at once if passing JSON or XML to module.
version_added: "2.5"
author:
    - "sookido (@sookido)"
    - "Logan Vig (@logan2211)"
    - "Dusan Matejka (@D3DeFi)"
requirements:
    - "python >= 2.6"
    - "zabbix-api >= 0.5.3"
options:
    template_name:
        description:
            - Name of Zabbix template.
            - Required when I(template_json) or I(template_xml) are not used.
            - Mutually exclusive with I(template_json) and I(template_xml).
        required: false
    template_json:
        description:
            - JSON dump of templates to import.
            - Multiple templates can be imported this way.
            - Mutually exclusive with I(template_name) and I(template_xml).
        required: false
        type: json
    template_xml:
        description:
            - XML dump of templates to import.
            - Multiple templates can be imported this way.
            - You are advised to pass XML structure matching the structure used by your version of Zabbix server.
            - Custom XML structure can be imported as long as it is valid, but may not yield consistent idempotent
              results on subsequent runs.
            - Mutually exclusive with I(template_name) and I(template_json).
        required: false
        version_added: '2.9'
    template_groups:
        description:
            - List of host groups to add template to when template is created.
            - Replaces the current host groups the template belongs to if the template is already present.
            - Required when creating a new template with C(state=present) and I(template_name) is used.
              Not required when updating an existing template.
        required: false
        type: list
    link_templates:
        description:
            - List of template names to be linked to the template.
            - Templates that are not specified and are linked to the existing template will be only unlinked and not
              cleared from the template.
        required: false
        type: list
    clear_templates:
        description:
            - List of template names to be unlinked and cleared from the template.
            - This option is ignored if template is being created for the first time.
        required: false
        type: list
    macros:
        description:
            - List of user macros to create for the template.
            - Macros that are not specified and are present on the existing template will be replaced.
            - See examples on how to pass macros.
        required: false
        type: list
        suboptions:
            name:
                description:
                    - Name of the macro.
                    - Must be specified in {$NAME} format.
            value:
                description:
                    - Value of the macro.
    dump_format:
        description:
            - Format to use when dumping template with C(state=dump).
        required: false
        choices: [json, xml]
        default: "json"
        version_added: '2.9'
    state:
        description:
            - Required state of the template.
            - On C(state=present) template will be created/imported or updated depending if it is already present.
            - On C(state=dump) template content will get dumped into required format specified in I(dump_format).
            - On C(state=absent) template will be deleted.
        required: false
        choices: [present, absent, dump]
        default: "present"

extends_documentation_fragment:
    - zabbix
'''

EXAMPLES = '''
---
- name: Create a new Zabbix template linked to groups, macros and templates
  local_action:
    module: zabbix_template
    server_url: http://127.0.0.1
    login_user: username
    login_password: password
    template_name: ExampleHost
    template_groups:
      - Role
      - Role2
    link_templates:
      - Example template1
      - Example template2
    macros:
      - macro: '{$EXAMPLE_MACRO1}'
        value: 30000
      - macro: '{$EXAMPLE_MACRO2}'
        value: 3
      - macro: '{$EXAMPLE_MACRO3}'
        value: 'Example'
    state: present

- name: Unlink and clear templates from the existing Zabbix template
  local_action:
    module: zabbix_template
    server_url: http://127.0.0.1
    login_user: username
    login_password: password
    template_name: ExampleHost
    clear_templates:
      - Example template3
      - Example template4
    state: present

- name: Import Zabbix templates from JSON
  local_action:
    module: zabbix_template
    server_url: http://127.0.0.1
    login_user: username
    login_password: password
    template_json: "{{ lookup('file', 'zabbix_apache2.json') }}"
    state: present

- name: Import Zabbix templates from XML
  local_action:
    module: zabbix_template
    server_url: http://127.0.0.1
    login_user: username
    login_password: password
    template_xml: "{{ lookup('file', 'zabbix_apache2.json') }}"
    state: present

- name: Import Zabbix template from Ansible dict variable
  zabbix_template:
    login_user: username
    login_password: password
    server_url: http://127.0.0.1
    template_json:
      zabbix_export:
        version: '3.2'
        templates:
          - name: Template for Testing
            description: 'Testing template import'
            template: Test Template
            groups:
              - name: Templates
            applications:
              - name: Test Application
    state: present

- name: Configure macros on the existing Zabbix template
  local_action:
    module: zabbix_template
    server_url: http://127.0.0.1
    login_user: username
    login_password: password
    template_name: Template
    macros:
      - macro: '{$TEST_MACRO}'
        value: 'Example'
    state: present

- name: Delete Zabbix template
  local_action:
    module: zabbix_template
    server_url: http://127.0.0.1
    login_user: username
    login_password: password
    template_name: Template
    state: absent

- name: Dump Zabbix template as JSON
  local_action:
    module: zabbix_template
    server_url: http://127.0.0.1
    login_user: username
    login_password: password
    template_name: Template
    state: dump
  register: template_dump

- name: Dump Zabbix template as XML
  local_action:
    module: zabbix_template
    server_url: http://127.0.0.1
    login_user: username
    login_password: password
    template_name: Template
    dump_format: xml
    state: dump
  register: template_dump
'''

RETURN = '''
---
template_json:
  description: The JSON dump of the template
  returned: when state is dump
  type: str
  sample: {
        "zabbix_export":{
            "date":"2017-11-29T16:37:24Z",
            "templates":[{
                "templates":[],
                "description":"",
                "httptests":[],
                "screens":[],
                "applications":[],
                "discovery_rules":[],
                "groups":[{"name":"Templates"}],
                "name":"Test Template",
                "items":[],
                "macros":[],
                "template":"test"
            }],
            "version":"3.2",
            "groups":[{
                "name":"Templates"
            }]
        }
    }

template_xml:
  description: dump of the template in XML representation
  returned: when state is dump and dump_format is xml
  type: str
  sample: |-
    <?xml version="1.0" ?>
    <zabbix_export>
        <version>4.2</version>
        <date>2019-07-12T13:37:26Z</date>
        <groups>
            <group>
                <name>Templates</name>
            </group>
        </groups>
        <templates>
            <template>
                <template>test</template>
                <name>Test Template</name>
                <description/>
                <groups>
                    <group>
                        <name>Templates</name>
                    </group>
                </groups>
                <applications/>
                <items/>
                <discovery_rules/>
                <httptests/>
                <macros/>
                <templates/>
                <screens/>
                <tags/>
            </template>
        </templates>
    </zabbix_export>
'''


import atexit
import json
import traceback
import xml.etree.ElementTree as ET

from distutils.version import LooseVersion
from ansible.module_utils.basic import AnsibleModule, missing_required_lib
from ansible.module_utils._text import to_native

try:
    from zabbix_api import ZabbixAPI, ZabbixAPIException

    HAS_ZABBIX_API = True
except ImportError:
    ZBX_IMP_ERR = traceback.format_exc()
    HAS_ZABBIX_API = False


class Template(object):
    def __init__(self, module, zbx):
        self._module = module
        self._zapi = zbx

    # check if host group exists
    def check_host_group_exist(self, group_names):
        for group_name in group_names:
            result = self._zapi.hostgroup.get({'filter': {'name': group_name}})
            if not result:
                self._module.fail_json(msg="Hostgroup not found: %s" %
                                       group_name)
        return True

    # get group ids by group names
    def get_group_ids_by_group_names(self, group_names):
        group_ids = []
        if group_names is None or len(group_names) == 0:
            return group_ids
        if self.check_host_group_exist(group_names):
            group_list = self._zapi.hostgroup.get(
                {'output': 'extend',
                 'filter': {'name': group_names}})
            for group in group_list:
                group_id = group['groupid']
                group_ids.append({'groupid': group_id})
        return group_ids

    def get_template_ids(self, template_list):
        template_ids = []
        if template_list is None or len(template_list) == 0:
            return template_ids
        for template in template_list:
            template_list = self._zapi.template.get(
                {'output': 'extend',
                 'filter': {'host': template}})
            if len(template_list) < 1:
                continue
            else:
                template_id = template_list[0]['templateid']
                template_ids.append(template_id)
        return template_ids

    def add_template(self, template_name, group_ids, link_template_ids, macros):
        if self._module.check_mode:
            self._module.exit_json(changed=True)

        self._zapi.template.create({'host': template_name, 'groups': group_ids, 'templates': link_template_ids,
                                    'macros': macros})

    def check_template_changed(self, template_ids, template_groups, link_templates, clear_templates,
                               template_macros, template_content, template_type):
        """Compares template parameters to already existing values if any are found.

        template_json - JSON structures are compared as deep sorted dictionaries,
        template_xml - XML structures are compared as strings, but filtered and formatted first,
        If none above is used, all the other arguments are compared to their existing counterparts
        retrieved from Zabbix API."""
        changed = False
        # Compare filtered and formatted XMLs strings for any changes. It is expected that provided
        # XML has same structure as Zabbix uses (e.g. it was optimally exported via Zabbix GUI or API)
        if template_content is not None and template_type == 'xml':
            existing_template = self.dump_template(template_ids, template_type='xml')

            if self.filter_xml_template(template_content) != self.filter_xml_template(existing_template):
                self._module.fail_json(msg='changed :\n%s\n\n\n%s' % (self.filter_xml_template(template_content), self.filter_xml_template(existing_template)))
                changed = True

            return changed

        existing_template = self.dump_template(template_ids, template_type='json')
        # Compare JSON objects as deep sorted python dictionaries
        if template_content is not None and template_type == 'json':
            parsed_template_json = self.load_json_template(template_content)
            if self.diff_template(parsed_template_json, existing_template):
                changed = True

            return changed

        # If neither template_json or template_xml were used, user provided all parameters via module options
        if template_groups is not None:
            existing_groups = [g['name'] for g in existing_template['zabbix_export']['groups']]

            if set(template_groups) != set(existing_groups):
                changed = True

        # Check if any new templates would be linked or any existing would be unlinked
        exist_child_templates = [t['name'] for t in existing_template['zabbix_export']['templates'][0]['templates']]
        if link_templates is not None:
            if set(link_templates) != set(exist_child_templates):
                changed = True

        # Mark that there will be changes when at least one existing template will be unlinked
        if clear_templates is not None:
            for t in clear_templates:
                if t in exist_child_templates:
                    changed = True
                    break

        if template_macros is not None:
            existing_macros = existing_template['zabbix_export']['templates'][0].get('macros')
            if template_macros != existing_macros:
                changed = True

        return changed

    def update_template(self, template_ids, group_ids, link_template_ids, clear_template_ids, template_macros):
        template_changes = {}
        if group_ids is not None:
            template_changes.update({'groups': group_ids})

        if link_template_ids is not None:
            template_changes.update({'templates': link_template_ids})

        if clear_template_ids is not None:
            template_changes.update({'templates_clear': clear_template_ids})

        if template_macros is not None:
            template_changes.update({'macros': template_macros})

        if template_changes:
            # If we got here we know that only one template was provided via template_name
            template_changes.update({'templateid': template_ids[0]})
            self._zapi.template.update(template_changes)

    def delete_template(self, templateids):
        if self._module.check_mode:
            self._module.exit_json(changed=True)
        self._zapi.template.delete(templateids)

    def ordered_json(self, obj):
        # Deep sort json dicts for comparison
        if isinstance(obj, dict):
            return sorted((k, self.ordered_json(v)) for k, v in obj.items())
        if isinstance(obj, list):
            return sorted(self.ordered_json(x) for x in obj)
        else:
            return obj

    def dump_template(self, template_ids, template_type='json'):
        if self._module.check_mode:
            self._module.exit_json(changed=True)

        try:
            dump = self._zapi.configuration.export({'format': template_type, 'options': {'templates': template_ids}})
            if template_type == 'xml':
                return ET.tostring(ET.fromstring(dump.encode('utf-8')), encoding='utf-8').decode()
            else:
                return self.load_json_template(dump)

        except ZabbixAPIException as e:
            self._module.fail_json(msg='Unable to export template: %s' % e)

    def diff_template(self, template_json_a, template_json_b):
        # Compare 2 zabbix templates and return True if they differ.
        template_json_a = self.filter_template(template_json_a)
        template_json_b = self.filter_template(template_json_b)
        if self.ordered_json(template_json_a) == self.ordered_json(template_json_b):
            return False
        return True

    def filter_template(self, template_json):
        # Filter the template json to contain only the keys we will update
        keep_keys = set(['graphs', 'templates', 'triggers', 'value_maps'])
        unwanted_keys = set(template_json['zabbix_export']) - keep_keys
        for unwanted_key in unwanted_keys:
            del template_json['zabbix_export'][unwanted_key]

        # Versions older than 2.4 do not support description field within template
        desc_not_supported = False
        if LooseVersion(self._zapi.api_version()).version[:2] < LooseVersion('2.4').version:
            desc_not_supported = True

        # Filter empty attributes from template object to allow accurate comparison
        for template in template_json['zabbix_export']['templates']:
            for key in list(template.keys()):
                if not template[key] or (key == 'description' and desc_not_supported):
                    template.pop(key)

        return template_json

    def filter_xml_template(self, template_xml):
        """Filters out keys from XML template that may wary between exports (e.g date or version) and
        keys that are not imported via this module.

        It is advised that provided XML template exactly matches XML structure used by Zabbix"""
        # Strip last new line and convert string to ElementTree
        parsed_xml_root = self.load_xml_template(template_xml.strip())
        keep_keys = ['graphs', 'templates', 'triggers', 'value_maps']

        # Remove unwanted XML nodes
        for node in list(parsed_xml_root):
            if node.tag not in keep_keys:
                parsed_xml_root.remove(node)

        # Filter empty attributes from template objects to allow accurate comparison
        for template in list(parsed_xml_root.find('templates')):
            for element in list(template):
                if element.text is None and len(list(element)) == 0:
                    template.remove(element)

        # Filter new lines and indentation
        # xml_root_text = list(line.strip() for line in ET.tostring(parsed_xml_root).split('\n'))
        xml_root_text = list(line.strip() for line in ET.tostring(parsed_xml_root, encoding='utf8', method='xml').decode().split('\n'))
        return ''.join(xml_root_text)

    def load_json_template(self, template_json):
        try:
            return json.loads(template_json)
        except ValueError as e:
            self._module.fail_json(msg='Invalid JSON provided', details=to_native(e), exception=traceback.format_exc())

    def load_xml_template(self, template_xml):
        try:
            return ET.fromstring(template_xml)
        except ET.ParseError as e:
            self._module.fail_json(msg='Invalid XML provided:' + template_xml, details=to_native(e), exception=traceback.format_exc())

    def import_template(self, template_content, template_type='json'):
        # rules schema latest version
        update_rules = {
            'applications': {
                'createMissing': True,
                'deleteMissing': True
            },
            'discoveryRules': {
                'createMissing': True,
                'updateExisting': True,
                'deleteMissing': True
            },
            'graphs': {
                'createMissing': True,
                'updateExisting': True,
                'deleteMissing': True
            },
            'httptests': {
                'createMissing': True,
                'updateExisting': True,
                'deleteMissing': True
            },
            'items': {
                'createMissing': True,
                'updateExisting': True,
                'deleteMissing': True
            },
            'templates': {
                'createMissing': True,
                'updateExisting': True
            },
            'templateLinkage': {
                'createMissing': True
            },
            'templateScreens': {
                'createMissing': True,
                'updateExisting': True,
                'deleteMissing': True
            },
            'triggers': {
                'createMissing': True,
                'updateExisting': True,
                'deleteMissing': True
            },
            'valueMaps': {
                'createMissing': True,
                'updateExisting': True
            }
        }

        try:
            # old api version support here
            api_version = self._zapi.api_version()
            # updateExisting for application removed from zabbix api after 3.2
            if LooseVersion(api_version).version[:2] <= LooseVersion('3.2').version:
                update_rules['applications']['updateExisting'] = True

            import_data = {'format': template_type, 'source': template_content, 'rules': update_rules}
            self._zapi.configuration.import_(import_data)
        except ZabbixAPIException as e:
            self._module.fail_json(msg='Unable to import template', details=to_native(e),
                                   exception=traceback.format_exc())


def main():
    module = AnsibleModule(
        argument_spec=dict(
            server_url=dict(type='str', required=True, aliases=['url']),
            login_user=dict(type='str', required=True),
            login_password=dict(type='str', required=True, no_log=True),
            http_login_user=dict(type='str', required=False, default=None),
            http_login_password=dict(type='str', required=False, default=None, no_log=True),
            validate_certs=dict(type='bool', required=False, default=True),
            template_name=dict(type='str', required=False),
            template_json=dict(type='json', required=False),
            template_xml=dict(type='str', required=False),
            template_groups=dict(type='list', required=False),
            link_templates=dict(type='list', required=False),
            clear_templates=dict(type='list', required=False),
            macros=dict(type='list', required=False),
            dump_format=dict(type='str', required=False, default='json', choices=['json', 'xml']),
            state=dict(default="present", choices=['present', 'absent', 'dump']),
            timeout=dict(type='int', default=10)
        ),
        required_one_of=[
            ['template_name', 'template_json', 'template_xml']
        ],
        mutually_exclusive=[
            ['template_name', 'template_json', 'template_xml']
        ],
        required_if=[
            ['state', 'absent', ['template_name']],
            ['state', 'dump', ['template_name']]
        ],
        supports_check_mode=True
    )

    if not HAS_ZABBIX_API:
        module.fail_json(msg=missing_required_lib('zabbix-api', url='https://pypi.org/project/zabbix-api/'), exception=ZBX_IMP_ERR)

    server_url = module.params['server_url']
    login_user = module.params['login_user']
    login_password = module.params['login_password']
    http_login_user = module.params['http_login_user']
    http_login_password = module.params['http_login_password']
    validate_certs = module.params['validate_certs']
    template_name = module.params['template_name']
    template_json = module.params['template_json']
    template_xml = module.params['template_xml']
    template_groups = module.params['template_groups']
    link_templates = module.params['link_templates']
    clear_templates = module.params['clear_templates']
    template_macros = module.params['macros']
    dump_format = module.params['dump_format']
    state = module.params['state']
    timeout = module.params['timeout']

    zbx = None
    try:
        zbx = ZabbixAPI(server_url, timeout=timeout, user=http_login_user, passwd=http_login_password,
                        validate_certs=validate_certs)
        zbx.login(login_user, login_password)
        atexit.register(zbx.logout)
    except ZabbixAPIException as e:
        module.fail_json(msg="Failed to connect to Zabbix server: %s" % e)

    template = Template(module, zbx)

    # Identify template names for IDs retrieval
    # Template names are expected to reside in ['zabbix_export']['templates'][*]['template'] for both data types
    template_content, template_type = None, None
    if template_json is not None:
        template_type = 'json'
        template_content = template_json
        json_parsed = template.load_json_template(template_content)
        template_names = list(t['template'] for t in json_parsed['zabbix_export']['templates'])

    elif template_xml is not None:
        template_type = 'xml'
        template_content = template_xml
        xml_parsed = template.load_xml_template(template_content)
        template_names = list(t.find('template').text for t in list(xml_parsed.find('templates')))

    else:
        template_names = [template_name]

    template_ids = template.get_template_ids(template_names)

    if state == "absent":
        if not template_ids:
            module.exit_json(changed=False, msg="Template not found. No changed: %s" % template_name)

        template.delete_template(template_ids)
        module.exit_json(changed=True, result="Successfully deleted template %s" % template_name)

    elif state == "dump":
        if not template_ids:
            module.fail_json(msg='Template not found: %s' % template_name)

        if dump_format == 'json':
            module.exit_json(changed=False, template_json=template.dump_template(template_ids, template_type='json'))
        elif dump_format == 'xml':
            module.exit_json(changed=False, template_xml=template.dump_template(template_ids, template_type='xml'))

    elif state == "present":
        # Load all subelements for template that were provided by user
        group_ids = None
        if template_groups is not None:
            group_ids = template.get_group_ids_by_group_names(template_groups)

        link_template_ids = None
        if link_templates is not None:
            link_template_ids = template.get_template_ids(link_templates)

        clear_template_ids = None
        if clear_templates is not None:
            clear_template_ids = template.get_template_ids(clear_templates)

        if template_macros is not None:
            # Zabbix configuration.export does not differentiate python types (numbers are returned as strings)
            for macroitem in template_macros:
                for key in macroitem:
                    macroitem[key] = str(macroitem[key])

        if not template_ids:
            # Assume new templates are being added when no ID's were found
            if template_content is not None:
                template.import_template(template_content, template_type)
                module.exit_json(changed=True, result="Template import successful")

            else:
                if group_ids is None:
                    module.fail_json(msg='template_groups are required when creating a new Zabbix template')

                template.add_template(template_name, group_ids, link_template_ids, template_macros)
                module.exit_json(changed=True, result="Successfully added template: %s" % template_name)

        else:
            changed = template.check_template_changed(template_ids, template_groups, link_templates, clear_templates,
                                                      template_macros, template_content, template_type)

            if module.check_mode:
                module.exit_json(changed=changed)

            if changed:
                if template_type is not None:
                    template.import_template(template_content, template_type)
                else:
                    template.update_template(template_ids, group_ids, link_template_ids, clear_template_ids,
                                             template_macros)

            module.exit_json(changed=changed, result="Template successfully updated")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import sys
from powerdnsadmin import create_app
from powerdnsadmin.models.user import User
from powerdnsadmin.models.setting import Setting

app = create_app()

with app.app_context():
    user = User(
        username='admin',
        plain_text_password=sys.argv[1],
        firstname='Administrator',
        role_id='Administrator')
    user.create_local_user()
    setting = Setting()
    forward_records_allow_edit = setting.get('forward_records_allow_edit')
    forward_records_allow_edit['LUA'] = True
    setting.set('forward_records_allow_edit', forward_records_allow_edit)
    setting.set('pdns_api_url', 'http://powerdns:8081/')
    setting.set('pdns_api_key', sys.argv[1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2018 Procube Co., Ltd.

"""
hive inventory: dynamic inventory plugin for hive
"""

from ansible.plugins.inventory import BaseInventoryPlugin
from ansible.errors import AnsibleParserError
from ansible.parsing.yaml.objects import AnsibleMapping, AnsibleSequence, AnsibleUnicode
from ansible.module_utils.six import text_type
from hive_builder.hive import hiveContext
from ansible.utils.display import Display

display = Display()

DOCUMENTATION = r'''
  name: hive_services
  plugin_type: inventory
  short_description: Returns Inventory for hive volumes, networks, images, service
  description:
    - Generate volumes, networks, images, services from service defition yaml

  options:
    plugin:
      description: token that ensures this is a source file for the 'hive_services' plugin.
      required: True
      choices: ['hive_services']
    services:
      description: service definitions
      required: True
    networks:
      description: network definitions
      required: False
    available_on:
      description: list of stage which the service available on
'''

EXAMPLES = r'''
plugin: hive_inventory
stages:
- stage: private
  provider: vagrant
  separate_repository: False
  name: test
  subnets:
  - cidr: 192.168.0.96/29
'''

STAGES = ['private', 'staging', 'production']


class InventoryModule(BaseInventoryPlugin):

  NAME = 'hive_services'  # used internally by Ansible, it should match the file name but not required

  def __init__(self):
    super(InventoryModule, self).__init__()
    self.sites = []

  def verify_file(self, path):
    ''' return true/false if this is possibly a valid file for this plugin to consume '''
    valid = False
    if super(InventoryModule, self).verify_file(path):
      # base class verifies that file exists and is readable by current user
      if path.endswith(('.yaml', '.yml')):
        valid = True
    return valid

  def parse(self, inventory, loader, path, cache=True):
    super(InventoryModule, self).parse(inventory, loader, path, cache)
    self.context = hiveContext()
    self.context.setup(None)
    self._read_config_data(path)
    self.inventory.add_group('services')
    self.inventory.add_group('images')
    self.inventory.add_group('volumes')
    self.inventory.add_group('drbd_volumes')
    self.inventory.add_group('networks')
    available_on = self.get_option('available_on')
    if available_on is None:
      available_on = STAGES
    for s in available_on:
      if s not in STAGES:
        raise AnsibleParserError(f'"element of "available_on" must be one of {STAGES}, but {s}')
      self.inventory.add_group(s)
    services = self.get_option('services')
    if type(services) != AnsibleMapping:
      raise AnsibleParserError(f'"services" must be dict type, but {type(services)}')
    device_id_map = self.context.vars.get('drbd_device_map', {})
    published_port_map = self.context.vars.get('published_port_map', {})
    for name, options in services.items():
      service = Service(name, options, available_on)
      service.parse(inventory, device_id_map, published_port_map)
    self.context.set_persistent('drbd_device_map', device_id_map)
    self.context.set_persistent('published_port_map', published_port_map)
    self.context.save_persistent()
    networks = self.get_option('networks')
    no_default = True
    if networks is not None:
      if type(networks) != AnsibleMapping:
        raise AnsibleParserError(f'"networks" must be dict type')
      for name, options in networks.items():
        if name == 'hive_default_network':
          no_default = False
        network = Network(name, options, available_on)
        network.parse(inventory)
    if no_default:
      network = Network('hive_default_network', AnsibleMapping(), available_on)
      network.parse(inventory)


IMAGE_PARAMS = ['from', 'roles', 'env', 'stop_signal', 'user', 'working_dir', 'standalone', 'entrypoint',
                'command', 'privileged', 'expose']
SERVICE_PARAMS_COPY = ['backup_scripts', 'command', 'dns', 'endpoint_mode', 'entrypoint', 'environment',
                       'hosts', 'initialize_roles', 'labels', 'logging', 'mode', 'placement', 'replicas',
                       'restart_config', 'standalone', 'user', 'working_dir']
NETWORK_PARAMS = ['driver', 'ipam', 'driver_opts']
SERVICE_PARAMS = SERVICE_PARAMS_COPY + ['volumes', 'image', 'ports', 'available_on']
VOLUME_PARAMS = ['target', 'type', 'source', 'readonly']
VOLUME_PARAMS_DEF = ['driver', 'driver_opts', 'drbd']


class Network:
  def __init__(self, name, options, available_on):
    self.name = name
    self.options = options
    self.available_on = available_on

  def parse(self, inventory):
    inventory.add_host(self.name, group='networks')
    for s in self.available_on:
      inventory.add_host(self.name, group=s)
    if self.options is None:
      return
    if type(self.options) != AnsibleMapping:
      raise AnsibleParserError(f'value must be dict type in network {self.name}')
    for option_name, option_value in self.options.items():
      if option_name not in NETWORK_PARAMS:
        raise AnsibleParserError(f'unknown parameter {option_name} is specified in network in service {self.name}')
      inventory.set_variable(self.name, f'hive_{option_name}', option_value)


class Service:
  def __init__(self, name, options, available_on):
    self.name = name
    self.options = options
    self.available_on = available_on

  def parse(self, inventory, device_id_map, published_port_map):
    if type(self.options) != AnsibleMapping:
      raise AnsibleParserError(f'value must be dict type in service {self.name}')
    inventory.add_host(self.name, group='services')
    for option_name, option_value in self.options.items():
      if option_name not in SERVICE_PARAMS:
        raise AnsibleParserError(f'unknown parameter {option_name} is specified in service {self.name}')
      if option_name in SERVICE_PARAMS_COPY:
        inventory.set_variable(self.name, f'hive_{option_name}', option_value)
    if 'available_on' in self.options:
      self.available_on = self.options['available_on']
      for s in self.available_on:
        if s not in STAGES:
          raise AnsibleParserError(f'"element of "available_on" of service {self.name} must be one of {STAGES}, but {s}')
    for s in self.available_on:
      inventory.add_host(self.name, group=s)
    if 'volumes' in self.options or self.options.get('standalone', False):
      volumes_value = self.options.get('volumes', AnsibleSequence([]))
      if type(volumes_value) != AnsibleSequence:
        raise AnsibleParserError(f'"volumes" must be list type in service {self.name}, but {type(volumes_value)}')
      volumes = []
      if self.options.get('standalone', False):
        volumes = [
            {'source': '/sys/fs/cgroup', 'target': '/sys/fs/cgroup', 'readonly': True},
            {'source': '', 'target': '/run', 'type': 'tmpfs'},
            {'source': '', 'target': '/tmp', 'type': 'tmpfs'}
        ]
      for volume in volumes_value:
        if type(volume) == AnsibleUnicode:
          raise AnsibleParserError(f'we do not support short syntax {text_type(volume)} in volume at service {self.name}')
        elif type(volume) == AnsibleMapping:
          if 'source' not in volume or 'target' not in volume:
              raise AnsibleParserError(f'both "source" and "target" must be specified in volume at service {self.name}')
          if 'driver' in volume:
            # we prepare volume at build-volume phase
            prepared_volume = {'name': volume['source'], 'driver': volume['driver']}
            if 'driver_opts' in volume:
              prepared_volume['dirver_options'] = volume['driver_opts']
            self.add_volume(inventory, prepared_volume, 'volumes')
          if 'drbd' in volume:
            if 'driver' in volume:
              raise AnsibleParserError(f'both "driver" and "drbd" can not be specified in volume at service {self.name}')
            if 'device_id' in volume['drbd']:
              if volume['drbd']['device_id'] in device_id_map.values() and volume['drbd']['device_id'] != device_id_map.get(volume['source']):
                raise AnsibleParserError(
                    f'duplicate deviceid {volume["drbd"]["device_id"]} of volume {volume["source"]} in service {self.name}')
            else:
              if volume['source'] in device_id_map:
                  volume['drbd']['device_id'] = device_id_map[volume['source']]
              else:
                # assign new device id
                for device_id in range(1, 128):
                  if device_id not in device_id_map.values():
                    volume['drbd']['device_id'] = device_id
                    break
                if 'device_id' not in volume['drbd']:
                  raise AnsibleParserError(f'too many drbd volumes or garbage device_id are remain in "device_id_map" of persistent hive variable')
            device_id_map[text_type(volume['source'])] = volume['drbd']['device_id']
            self.add_volume(inventory, {'name': volume['source'], 'drbd': volume['drbd']}, 'drbd_volumes')
          volume_value = {}
          for k, v in volume.items():
            # TODO: support properties of mounts property of docker_swarm_service module
            # I do not know what driver_config option means, it does check for volume definition? or it does create volume ondemand?
            if k not in VOLUME_PARAMS + VOLUME_PARAMS_DEF:
              raise AnsibleParserError(f'unknown parameter {k} is specified in volume in service {self.name}')
            if k in VOLUME_PARAMS:
              volume_value[k] = v
          volumes.append(volume_value)
        else:
          raise AnsibleParserError(f'all element of "volumes" must be dict type or str type in service {self.name}')
      inventory.set_variable(self.name, 'hive_volumes', volumes)

    if 'image' in self.options:
      image_value = self.options['image']
      if type(image_value) == AnsibleUnicode:
        # self.options['image'] can be {{hive_registry}}/image_some_another_service to refer build image
        inventory.set_variable(self.name, 'hive_image', image_value)
      elif type(image_value) == AnsibleMapping:
        # this container needs build.
        image_name = f'image_{self.name}'
        inventory.add_host(image_name, group='images')
        for s in self.available_on:
          inventory.add_host(image_name, group=s)
        # when hive_image_name is set, means image = {{hive_registry}}:{{hive_image_name}}
        inventory.set_variable(self.name, 'hive_image_name', image_name)
        if 'from' not in image_value:
          raise AnsibleParserError(f'"from" must be specified in "image" at service {self.name}')
        # just copy paste parameter for build image
        if self.options.get('standalone', False):
          inventory.set_variable(image_name, f'hive_standalone', True)
          inventory.set_variable(image_name, f'hive_privileged', True)
        for option_name, option_value in image_value.items():
          if option_name not in IMAGE_PARAMS:
            raise AnsibleParserError(f'unknown parameter {option_name} is specified in "image" at service {self.name}')
          inventory.set_variable(image_name, f'hive_{option_name}', option_value)
      else:
        raise AnsibleParserError(f'"image" must be dict type or str type in service {self.name}, but type is {type(image_value)}')
    if 'ports' in self.options:
      ports = []
      for portdef in self.options['ports']:
        if type(portdef) == AnsibleMapping:
          pass
        elif type(portdef) == AnsibleUnicode:
          slash = portdef.split('/')
          protocol = 'tcp'
          if len(slash) > 1:
            protocol = slash[1]
          colon = slash[0].split(':')
          if len(colon) > 1:
            portdef = {'published_port': int(colon[0]), 'target_port': int(colon[1]), 'protocol': protocol}
          else:
            portdef = {'target_port': int(colon[0]), 'protocol': protocol}
        else:
          raise AnsibleParserError(f'element of "ports" must be dict type or str type in service {self.name}, but type is {type(portdef)}')
        if portdef['protocol'] not in ['tcp', 'udp', 'sctp']:
          raise AnsibleParserError(f'unknown protocol {portdef["protocol"]} is specified {self.name}.')
        published_port_key = self.name + text_type(portdef['target_port'])
        if 'published_port' in portdef:
          if portdef['published_port'] in published_port_map.values() and portdef['published_port'] != published_port_map.get(published_port_key):
            display.warning(
                f'duplicate port number {portdef["published_port"]} of port {published_port_key}')
        else:
          if published_port_key in published_port_map:
              portdef['published_port'] = published_port_map[published_port_key]
          else:
            # assign new port number
            for port_number in range(61001, 65535):
              if port_number not in published_port_map.values():
                portdef['published_port'] = port_number
                break
            if 'published_port' not in portdef:
              raise AnsibleParserError(f'too many published ports or garbage port number are remain in "published_port_map" of persistent hive variable')
        published_port_map[published_port_key] = portdef['published_port']
        ports.append(portdef)
      inventory.set_variable(self.name, f'hive_ports', ports)

  def add_volume(self, inventory, volume, group):
    name = 'volume_' + volume['name']
    inventory.add_host(name, group=group)
    for s in self.available_on:
      inventory.add_host(name, group=s)
    inventory.set_variable(name, 'hive_volume', volume)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2018 Procube Co., Ltd.

import docker
import os
import logging
import subprocess
import re
import ipaddress
import json
import signal
import time

logging.basicConfig(level=os.environ.get('HIVE_LOG_LEVEL', logging.INFO))
DAEMON = None


class HookBase:
  def __init__(self, label_value, serivce_id, service_name):
    self.label_value = label_value
    self.serivce_id = serivce_id
    self.service_name = service_name
    self.stay = False

  @property
  def id(self):
    return self.label_value + '@' + self.service_name + '#' + self.__class__.__name__

  @classmethod
  def check_service(cls, service):
    label_value = service.attrs.get('Spec', {}).get('Labels', {}).get(cls.label_name)
    if label_value is not None:
      DAEMON.logger.debug(f'label "{cls.label_name}" is found at service {service.name}')
      return cls.get_hook(label_value, service.id, service.name)
    return None

  @classmethod
  def get_hook(cls, label_value, serivce_id, name):
    return cls(label_value, serivce_id, name)

  def check_tasks(self):
    stay = False
    for task in DAEMON.client.services.get(self.serivce_id).tasks():
      if task.get('DesiredState') == 'running' and task.get('NodeID') == DAEMON.node_id:
        DAEMON.logger.debug(f'found task: {json.dumps(task)}')
        stay = True
    if self.stay and not stay:
      self.on_leave()
    elif stay and not self.stay:
      self.on_enter()
    self.stay = stay

  def on_registered(self):
    self.check_tasks()

  def on_unregistered(self):
    self.on_leave()

  def on_node_down(self):
    self.on_leave()


class MarkNode(HookBase):
  label_name = 'HIVE_MARK'

  def on_enter(self):
    node = DAEMON.get_node()
    if node is None:
      DAEMON.logger.info(f'fail to get node at on_enter, do nothing')
      return
    spec = node.attrs.get('Spec', {}).copy()
    labels = spec.get('Labels', {}).copy()
    if labels.get(self.label_value) == "true":
      DAEMON.logger.info(f'label "{self.label_value}" value is alredy set to "true" at on_enter, do nothing')
      return
    labels[self.label_value] = "true"
    spec['Labels'] = labels
    DAEMON.logger.debug(f'update node "{node.id}" Spec to "{json.dumps(spec)}"')
    node.update(spec)

  def on_leave(self):
    node = DAEMON.get_node()
    if node is None:
      DAEMON.logger.info(f'fail to get node at on_leave, do nothing')
      return
    spec = node.attrs.get('Spec', {}).copy()
    labels = spec.get('Labels', {}).copy()
    if self.label_value not in labels:
      DAEMON.logger.info(f'label "{self.label_value}" value is not set at on_leave, do nothing')
      return
    labels.pop(self.label_value)
    spec['Labels'] = labels
    DAEMON.logger.debug(f'update node "{node.id}" Spec to "{json.dumps(spec)}"')
    node.update(spec)


class SetVIP(HookBase):
  label_name = 'HIVE_VIP'

  @classmethod
  def check_service(cls, service):
    label_value = service.attrs.get('Spec', {}).get('Labels', {}).get(cls.label_name)
    if label_value is None:
      return None
    me = cls.get_hook(label_value, service.id, service.name)
    if me is None:
      return None
    me.router = service.attrs.get('Spec', {}).get('Labels', {}).get('HIVE_ROUTER')
    DAEMON.logger.debug(f'label "HIVE_ROUTER" value is {me.router}')
    return me

# sample of output of ip addr command
# 1: lo: <LOOPBACK,UP,LOWER_UP> mtu 65536 qdisc noqueue state UNKNOWN group default qlen 1000
#     link/loopback 00:00:00:00:00:00 brd 00:00:00:00:00:00
#     inet 127.0.0.1/8 scope host lo
#        valid_lft forever preferred_lft forever
#     inet 10.1.32.12/32 scope global lo
#        valid_lft forever preferred_lft forever
#     inet6 ::1/128 scope host
#        valid_lft forever preferred_lft forever
# 2: em1: <BROADCAST,MULTICAST,SLAVE,UP,LOWER_UP> mtu 1500 qdisc mq master bond0 state UP group default qlen 1000
#     link/ether 6c:2b:59:ab:5d:16 brd ff:ff:ff:ff:ff:ff
# 3: em2: <BROADCAST,MULTICAST,SLAVE,UP,LOWER_UP> mtu 1500 qdisc mq master bond0 state UP group default qlen 1000
#     link/ether 6c:2b:59:ab:5d:16 brd ff:ff:ff:ff:ff:ff
# 4: bond0: <BROADCAST,MULTICAST,MASTER,UP,LOWER_UP> mtu 1500 qdisc noqueue state UP group default qlen 1000
#     link/ether 6c:2b:59:ab:5d:16 brd ff:ff:ff:ff:ff:ff
#     inet 10.1.32.2/20 brd 10.1.47.255 scope global noprefixroute bond0
#        valid_lft forever preferred_lft forever
#     inet 10.1.32.12/20 scope global secondary bond0
#        valid_lft forever preferred_lft forever
#     inet6 fe80::6e2b:59ff:feab:5d16/64 scope link
#        valid_lft forever preferred_lft forever
# 5: docker_gwbridge: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc noqueue state UP group default
#     link/ether 02:42:0f:41:6f:78 brd ff:ff:ff:ff:ff:ff
#     inet 172.21.34.1/25 brd 172.21.34.127 scope global docker_gwbridge
#        valid_lft forever preferred_lft forever
#     inet6 fe80::42:fff:fe41:6f78/64 scope link
#        valid_lft forever preferred_lft forever

  reg_start_if = re.compile(r'^ *(\d+): *([\w@]+): .*$')
  reg_link = re.compile(r'^ +link/.*$')
  reg_inet = re.compile(r'^ +inet (\d+\.\d+\.\d+\.\d+/\d+) .*$')

  def subprocess_run(self, args):
    proc = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='UTF-8')
    if proc.returncode != 0:
      DAEMON.logger.error(f'fail to execute "{" ".join(args)}" command. stderr: {proc.stderr}')
      return proc
    DAEMON.logger.info(f'success to execute "{" ".join(args)}" command. stdout: {proc.stdout}')
    return proc

  def list_ifs(self):
    ip_addr = self.subprocess_run(['ip', 'addr'])
    state = 'start'
    interface = {}
    for line in ip_addr.stdout.splitlines():
      if state == 'start':
        start_if = self.__class__.reg_start_if.match(line)
        assert start_if, f'fail to parse line {line} does not match regexp {self.__class__.reg_start_if.pattern} in {state} state.'
        state = 'in_if'
        interface['name'] = start_if.group(2)
      elif state == 'in_if':
        if self.__class__.reg_link.match(line) is None:
          inet = self.__class__.reg_inet.match(line)
          if inet:
            inet_if = ipaddress.ip_interface(inet.group(1))
            interface['network'] = inet_if.network
            if 'ips' not in interface:
              interface['ips'] = []
            interface['ips'].append(inet_if.ip)
          else:
            start_if = self.__class__.reg_start_if.match(line)
            if start_if:
              if 'name' in interface and 'network' in interface:
                yield interface
              interface = dict(name=start_if.group(2))
    if 'name' in interface and 'network' in interface:
      yield interface

  def get_interface(self):
    found = None
    vip = ipaddress.ip_address(self.label_value)
    for interface in self.list_ifs():
      if vip in interface['network']:
        found = interface
        found['vip_if'] = ipaddress.ip_interface(str(vip) + '/' + str(interface['network'].prefixlen))
        break
    return found

  def on_enter(self):
    interface = self.get_interface()
    if interface is None:
      DAEMON.logger.error(f'No network interface to which a virtual IP {self.label_value} can be added was found.')
      return
    if interface['vip_if'].ip in interface['ips']:
      DAEMON.logger.info(f'vip {str(interface["vip_if"].ip)} is already bound on interface {interface["name"]} at on_enter')
      return
    self.setVip(interface)

  def setVip(self, interface):
    self.subprocess_run(['ip', 'addr', 'add', interface['vip_if'].with_prefixlen, 'dev', interface['name']])
    self.subprocess_run(['arping', '-c', '1', '-A', '-I', interface['name'], str(interface['vip_if'].ip)])
    if self.router:
      rest_count = 5
      while rest_count > 0:
        ping_proc = self.subprocess_run(['ping', '-c', '1', self.router, '-I', str(interface['vip_if'].ip)])
        if ping_proc.returncode == 0:
          return
        rest_count -= 1
        time.sleep(10)
      DAEMON.logger.error(f'Exceed max retry ping test count:5')
      self.clearVip(interface)

  def clearVip(self, interface):
    self.subprocess_run(['ip', 'addr', 'del', interface['vip_if'].with_prefixlen, 'dev', interface['name']])

  def on_leave(self):
    interface = self.get_interface()
    if interface is None:
      DAEMON.logger.error(f'any network interface is not found for vip {self.label_name}')
      return
    if interface['vip_if'].ip not in interface['ips']:
      DAEMON.logger.info(f'vip {str(interface["vip_if"].ip)} is not bound on interface {interface["name"]} at on_leave')
      return
    self.clearVip(interface)


class FollowSwarmServiceDaemon:
  def __init__(self):
    self.client = docker.from_env()
    client_info = self.client.info()
    self.node_name = client_info.get("Name")
    self.node_id = client_info.get('Swarm', {}).get('NodeID')
    self.logger = logging.getLogger(f'FollowSwarm@{self.node_name}')
    self.logger.debug(f'client.info():{json.dumps(client_info)}')
    self.hooks = {}
    self.hook_classes = [SetVIP, MarkNode]

  def list_hooks(self):
    for service in self.client.services.list():
      self.logger.debug(f'check if the service to be caputured: {service.name}')
      for hook_class in self.hook_classes:
        hook = hook_class.check_service(service)
        if hook is not None:
          yield hook

  def check_services(self):
    deleted_hooks = self.hooks.copy()
    new_hooks = []
    for hook in self.list_hooks():
      self.logger.debug(f'find a hook: {hook.id}')
      if hook.id in deleted_hooks:
        deleted_hooks.pop(hook.id)
      else:
        new_hooks.append(hook)
    for hook in deleted_hooks.values():
      self.logger.debug(f'{hook.id} is not found, so unregister it.')
      self.hooks.pop(hook.id)
      hook.on_unregistered()
    for hook in new_hooks:
      self.logger.debug(f'{hook.id} is new, so register it.')
      self.hooks[hook.id] = hook
      hook.on_registered()

  def check_tasks(self):
    for hook in self.hooks.values():
      hook.check_tasks()

  def hook_node_down(self):
    self.logger.info('shutdown process started')
    try:
      for hook in self.hooks.values():
        hook.on_node_down()
    except Exception as e:
      self.logger.exception(f'exception occur in shutdown process: {e}')

  def run(self):
    self.logger.info('Follow Swarm Service Daemon Started')
    self.check_services()
    try:
      for ev in self.client.events(decode=True):
        if ev.get('Type') == 'service':
          self.logger.debug('service type event received')
          self.check_services()
          self.logger.debug('container type event received')
        if ev.get('Type') == 'container':
          self.check_tasks()
      # when events() return End of Element, it means node is down
    except docker.errors.APIError as e:
      # raise exception from docker api, it means node is down
      self.logger.exception(f'fail to call docker api: {e}')
    except AssertionError as e:
      self.logger.exception(f'unexpected result: {e}')
    except KeyboardInterrupt:
      self.logger.info(f'keyboard interrupted')
    except SystemExit:
      self.logger.info(f'receive terminate signal')
    self.hook_node_down()
    self.logger.info('Follow Swarm Service Daemon Stopped')

  def get_node(self):
    node = None
    try:
      node = self.client.nodes.get(self.node_id)
    except Exception as e:
      self.logger.exception(f'fail to get node for {self.node_name}: %s', e)
    return node


def on_terminate(signum, frame):
  raise SystemExit('Signal received')


def main():
  global DAEMON
  signal.signal(signal.SIGTERM, on_terminate)
  try:
    DAEMON = FollowSwarmServiceDaemon()
  except docker.errors.APIError as e:
    logging.getLogger('FollowSwarm@UNKNOWN').error(f'fail to initialize docker client: {e}')
  DAEMON.run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# this module is downloaded from https://raw.githubusercontent.com/robparrott/ansible-vagrant/master/library/vagrant
#  2019/06/17 Mitsuru Nakakawaji <mitsuru@procube.jp>

# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from ansible.module_utils.basic import AnsibleModule
import sys
import subprocess
import os.path

DOCUMENTATION = '''
---
module: vagrant
short_description: create a local instance via vagrant
description:
     - creates VM instances via vagrant and optionally waits for it to be 'running'. This module has a dependency on python-vagrant.
version_added: "100.0"
options:
  state:
    description: Should the VMs be "present" or "absent."
  cmd:
    description:
      - vagrant subcommand to execute. Can be "up," "status," "config," "ssh," "halt," "destroy" or "clear."
    required: false
    default: null
    aliases: ['command']
  box_name:
    description:
      - vagrant boxed image to start
    required: false
    default: null
    aliases: ['image']
  box_path:
    description:
      - path to vagrant boxed image to start
    required: false
    default: null
    aliases: []
  instaces:
    description:
      - list of instance definition dict has inventory_hostname, private_ip
    required: True
  forward_ports:
    description:
      - comma separated list of ports to forward to the host
    required: False
    aliases: []
  vagrant_dir:
    description:
      - directory where vagrant command is executed
    required: True
  vm_name:
    description:
      - name of VM
    required: false
    default: null
    aliases: []

examples:
   - code: 'local_action: vagrant cmd=up box_name=lucid32 vm_name=webserver'
     description:
requirements: [ "vagrant" ]
author: Rob Parrott
'''

VAGRANT_FILE = "./Vagrantfile"
VAGRANT_DICT_FILE = "./Vagrantfile.json"
VAGRANT_LOCKFILE = "./.vagrant-lock"

VAGRANT_FILE_HEAD = "Vagrant::Config.run do |config|\n"
VAGRANT_FILE_BOX_NAME = "  config.vm.box = \"%s\"\n"
VAGRANT_FILE_VM_STANZA_HEAD = """
  config.vm.define :%s do |%s_config|
    %s_config.vm.network :hostonly, "%s"
    %s_config.vm.box = "%s"
"""
VAGRANT_FILE_HOSTNAME_LINE = "    %s_config.vm.host_name = \"%s\"\n"
VAGRANT_FILE_PORT_FORWARD_LINE = "    %s_config.vm.forward_port %s, %s\n"
VAGRANT_FILE_VM_STANZA_TAIL = "  end\n"

VAGRANT_FILE_TAIL = "\nend\n"

# If this is already a network on your machine, this may fail ... change it here.
VAGRANT_INT_IP = "192.168.179.%s"

DEFAULT_VM_NAME = "ansible"


try:
  import lockfile
except ModuleNotFoundError:
  print('Python module lockfile is not installed. Falling back to using flock(), which will fail on Windows.')

try:
  import vagrant
except ModuleNotFoundError:
  print("failed=True msg='python-vagrant required for this module'")
  sys.exit(1)


class VagrantWrapper(object):

  def __init__(self):
    '''
    Wrapper around the python-vagrant module for use with ansible.
    Note that Vagrant itself is non-thread safe, as is the python-vagrant lib, so we need to lock on basically all operations ...
    '''
    # Get a lock
    self.lock = None

    try:
      self.lock = lockfile.FileLock(VAGRANT_LOCKFILE)
      self.lock.acquire()
    except Exception:
      # fall back to using flock instead ...
      try:
        import fcntl
        self.lock = open(VAGRANT_LOCKFILE, 'w')
        fcntl.flock(self.lock, fcntl.LOCK_EX)
      except Exception:
        print("failed=True msg='Could not get a lock for using vagrant. Install python module \"lockfile\" to use vagrant on non-POSIX filesytems.'")
        sys.exit(1)

    # Initialize vagrant and state files
    self.vg = vagrant.Vagrant()

    # operation will create a default data structure if none present
    self.vg_data = {"instances": {}, "num_inst": 0}

  def __del__(self):
    "Clean up file locks"
    try:
      self.lock.release()
    except Exception:
      os.close(self.lock)
      os.unlink(self.lock)

  def prepare_box(self, box_name, box_path):
    """
    Given a specified name and URL, import a Vagrant "box" for use.
    """
    changed = False
    if box_name is None:
      raise Exception("You must specify a box_name with a box_path for vagrant.")
    boxes = self.vg.box_list()
    if box_name not in boxes:
      self.vg.box_add(box_name, box_path)
      changed = True

    return changed

  def build_instances(self, box_name, instances, box_path=None, ports=[]):
    """
    Fire up a given VM and name it, using vagrant's multi-VM mode.
    """
    changed = False
    if box_name is None:
      raise Exception("You must specify a box name for Vagrant.")
    if box_path is not None and self.prepare_box(box_name, box_path):
      changed = True

    for instance in instances:

      d = self._get_instance(instance['inventory_hostname'], instance['hive_private_ip'])
      if 'box_name' not in d:
        d['box_name'] = box_name
      d['forward_ports'] = ports

      # Save our changes and run
      self._instances()[instance['inventory_hostname']] = d

    return changed

  def up(self, check_mode, vm_name=None):
    changed = False

    vm_names = []
    if vm_name is not None:
      vm_names = [vm_name]
    else:
      vm_names = self._instances().keys()
    statuses = {}
    for vmn in vm_names:
      if self._get_status(vmn) != 'running':
        if not check_mode:
          self.vg.up(vmn, provision=True)
        changed = True
      statuses[vmn] = self._get_status(vmn)

    # ad = self._build_instance_array_for_ansible()
    return (changed, statuses)

  def status(self, vm_name=None):
    """
    Return the run status of the VM instance. If no instance N is given, returns first instance.
    """
    vm_names = []
    if vm_name is not None:
      vm_names = [vm_name]
    else:
      vm_names = self._instances().keys()

    statuses = {}
    for vmn in vm_names:
      statuses[vmn] = self._get_status(vmn)

    return (False, statuses)

  def config(self, vm_name=None):
    """
    Return info on SSH for the running instance.
    """
    vm_names = []
    if vm_name is not None:
      vm_names = [vm_name]
    else:
      vm_names = self._instances().keys()

    configs = {}
    for vmn in vm_names:
      cnf = self.vg.conf(None, vmn)
      configs[vmn] = cnf

    return (False, configs)

  def halt(self, check_mode, vm_name=None):
    """
    Shuts down a vm_name or all VMs.
    """
    changed = False
    vm_names = []
    if vm_name is not None:
      vm_names = [vm_name]
    else:
      vm_names = self._instances().keys()

    statuses = {}
    for vmn in vm_names:
      if self._get_status(vmn) == 'running':
        if not check_mode:
          self.vg.halt(vmn)
        changed = True
      statuses[vmn] = self._get_status(vmn)[0]

    return (changed, statuses)

  def destroy(self, check_mode, vm_name=None):
    """
    Halt and remove data for a VM, or all VMs.
    """
    changed = False

    vm_names = []
    if vm_name is not None:
      vm_names = [vm_name]
    else:
      vm_names = self._instances().keys()

    statuses = {}
    for vmn in vm_names:
      if not self._get_status(vmn) in ['absent', 'not_created']:
        if not check_mode:
          self.vg.destroy(vmn)
        changed = True
      statuses[vmn] = self._get_status(vmn)

    return (changed, statuses)

  def clear(self, check_mode, vm_name=None):
    """
    Halt and remove data for a VM, or all VMs. Also clear all state data
    """
    changed = self.vg.destroy(check_mode, vm_name)

    if os.path.isfile(VAGRANT_FILE):
      if not check_mode:
        os.remove(VAGRANT_FILE)
      changed = True
    if os.path.isfile(VAGRANT_DICT_FILE):
      if not check_mode:
        os.remove(VAGRANT_DICT_FILE)
      changed = True

    return changed

#
# Helper Methods
#

  def _instances(self):
    return self.vg_data['instances']

  def _get_instance(self, vm_name, ip):

    instances = self._instances()

    if vm_name in instances:
      return instances[vm_name]

    #
    # otherwise create one afresh
    #

    d = dict()
    N = self.vg_data['num_inst'] + 1
    # n = len(instances.keys())+1
    d['N'] = N
    d['name'] = vm_name
    d['internal_ip'] = ip
    d['forward_ports'] = []
    self.vg_data['num_inst'] = N

    self._instances()[vm_name] = d

    return d

  def _get_status(self, name):
    state = 'absent'
    try:
      state = self.vg.status(name)[0].state
    except Exception:
      pass
    return state

  #
  # Translate the state dictionary into the Vagrantfile
  #
  def write_vagrantfile(self):

    vfile = open(VAGRANT_FILE, 'w')
    vfile.write(VAGRANT_FILE_HEAD)

    instances = self._instances()
    for vm_name in instances.keys():
      d = instances[vm_name]
      name = d['name']
      ip = d['internal_ip']
      box_name = d['box_name']
      vfile.write(VAGRANT_FILE_VM_STANZA_HEAD %
                  (name, name, name, ip, name, box_name))
      vfile.write(VAGRANT_FILE_HOSTNAME_LINE % (name, name.replace('_', '-')))
      if 'forward_ports' in d:
        for p in d['forward_ports']:
          vfile.write(VAGRANT_FILE_PORT_FORWARD_LINE % (name, p, p))
      vfile.write(VAGRANT_FILE_VM_STANZA_TAIL)

    vfile.write(VAGRANT_FILE_TAIL)
    vfile.close()

  #
  # To be returned to ansible with info about instances
  #
  def _build_instance_array_for_ansible(self, vmname=None):

    vm_names = []
    instances = self._instances()
    if vmname is not None:
      vm_names = [vmname]
    else:
      vm_names = instances.keys()

    ans_instances = []
    for vm_name in vm_names:
      cnf = self.vg.conf(None, vm_name)
      if cnf is not None:
        d = {
            'name': vm_name,
            'id': cnf['Host'],
            'public_ip': cnf['HostName'],
            'internal_ip': instances[vm_name]['internal_ip'],
            'public_dns_name': cnf['HostName'],
            'port': cnf['Port'],
            'username': cnf['User'],
            'key': cnf['IdentityFile'],
            'status': self._get_status(vm_name)
        }
        ans_instances.append(d)

    return ans_instances

# --------
# MAIN
# --------


def main():

  module = AnsibleModule(
      argument_spec=dict(
          state=dict(),
          cmd=dict(required=False, aliases=['command']),
          box_name=dict(required=False, aliases=['image']),
          box_path=dict(),
          instances=dict(required=True, type='list', elements='dict'),
          forward_ports=dict(),
          vagrant_dir=dict(required=True)
      ),
      supports_check_mode=True
  )

  state = module.params.get('state')
  cmd = module.params.get('cmd')
  box_name = module.params.get('box_name')
  box_path = module.params.get('box_path')
  forward_ports = module.params.get('forward_ports')
  vagrant_dir = module.params.get('vagrant_dir')
  vm_name = module.params.get('vm_name')
  changed = False

  if forward_ports is not None:
    forward_ports = forward_ports.split(',')
  if forward_ports is None:
    forward_ports = []

  instances = module.params.get('instances')

  #
  # Check if we are being invoked under an idempotency idiom of "state=present" or "state=absent"
  #
  try:
    # Initialize vagrant
    os.chdir(vagrant_dir)
    vgw = VagrantWrapper()
    if vgw.build_instances(box_name, instances, box_path, forward_ports):
      changed = True

    if state is not None:

      if state != 'present' and state != 'absent':
        module.fail_json(msg="State must be \"present\" or \"absent\" in vagrant module.")

      if state == 'present':

        # vgw.write_vagrantfile()
        (changed, status) = vgw.up(module.check_mode)
        module.exit_json(changed=changed, status=status)

      if state == 'absent':
        (changed, status) = vgw.halt()
        module.exit_json(changed=changed, status=status)
    else:
      if cmd is None:
        module.fail_json(msg="either cmd or state must be specified in vagrant module.")

      if cmd == 'up':

        # vgw.write_vagrantfile()
        (changed, status) = vgw.up(module.check_mode)
        module.exit_json(changed=changed, status=status)

      elif cmd == 'status':

        # if vm_name == None:
        #    module.fail_json(msg="Error: you must specify a vm_name when calling status." )

        (changed, result) = vgw.status()
        module.exit_json(changed=changed, status=result)

      elif cmd == "config" or cmd == "conf":

        (changed, cnf) = vgw.config()
        module.exit_json(changed=changed, config=cnf)

      elif cmd == 'ssh':

        if vm_name is None:
          module.fail_json(msg="Error: you must specify a vm_name when calling ssh.")

        (changed, cnf) = vgw.config(vm_name)
        sshcmd = "ssh -i %s -p %s %s@%s" % (cnf["IdentityFile"], cnf["Port"], cnf["User"], cnf["HostName"])
        sshmsg = "Execute the command \"vagrant ssh %s\"" % (vm_name)
        module.exit_json(changed=changed, msg=sshmsg, SshCommand=sshcmd)

#            elif cmd == "load_key":
#
#                if vm_name == None:
#                    module.fail_json(msg="Error: you must specify a vm_name when calling load_key." )
#
#                cnf = vg.config(vm_name)
#                keyfile=cnf["IdentityFile"]
#
#                # Get loaded keys ...
#                loaded_keys = subprocess.check_output(["ssh-add", "-l"])
#                module.exit_json(changed = True, msg = loaded_keys)
#
#                subprocess.call(["ssh-add", keyfile])
#
#                module.exit_json(changed = True, msg = sshmsg, SshCommand = sshcmd)

      elif cmd == 'halt':

        (changed, status) = vgw.halt(module.check_mode)
        module.exit_json(changed=changed, status=status)

      elif cmd == 'destroy':

        (changed, status) = vgw.destroy(module.check_mode)
        module.exit_json(changed=changed, status=status)

      elif cmd == 'clear':

        changed = vgw.clear()
        module.exit_json(changed=changed)

      else:

        module.fail_json(msg="Unknown vagrant subcommand: \"%s\"." % (cmd))

  except subprocess.CalledProcessError as e:
    module.fail_json(msg="Vagrant command failed: %s. cwd=%s" % (e, os.getcwd()))
#    except Exception as e:
#        module.fail_json(msg=e.__str__())
  module.exit_json(status="success")


# this is magic, see lib/ansible/module_common.py
# <<INCLUDE_ANSIBLE_MODULE_COMMON>>

main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

DOCUMENTATION = '''
---
module: powerdns_record
short_description: Manage PowerDNS records
description:
- Create, update or delete a PowerDNS records using API
- A record is unique identified by name and type
- Changing a records type is therefore not possible
options:
  content:
    description:
    - Content of the record
    - Could be an ip address or hostname
  name:
    description:
    - Record name
    - If name is not an FQDN, zone will be added at the end to create an FQDN
    required: true
  server:
    description:
    - Server name.
    required: false
    default: localhost
  ttl:
    description:
    - Record TTL
    required: false
    default: 86400
  type:
    description:
    - Record type
    required: false
    choices: ['A', 'AAAA', 'CNAME', 'MX', 'PTR', 'SOA', 'SRV', 'TXT', 'LUA']
    default: None
  zone:
    description:
    - Name of zone where to ensure the record
    required: true
  pdns_host:
    description:
    - Name or ip address of PowerDNS host
    required: false
    default: 127.0.0.1
  pdns_port:
    description:
    - Port used by PowerDNS API
    required: false
    default: 8081
  pdns_prot:
    description:
    - Protocol used by PowerDNS API
    required: false
    default: http
    choices: ['http', 'https']
  pdns_api_key:
    description:
    - API Key to authenticate through PowerDNS API
  strict_ssl_checking:
    description:
    - Disables strict certificate checking
    default: false
author: "Thomas Krahn (@nosmoht)"
'''

EXAMPLES = '''
- powerdns_record:
    name: host01.internal.example.com
    type: A
    content: 192.168.1.234
    state: present
    zone: internal.example.com
    pdns_host: powerdns.example.cm
    pdns_port: 8080
    pdns_prot: http
    pdns_api_key: topsecret
'''

import requests # noqa


class PowerDNSError(Exception):
    def __init__(self, url, status_code, message):
        self.url = url
        self.status_code = status_code
        self.message = message
        super(PowerDNSError, self).__init__()


class PowerDNSClient:
    def __init__(self, host, port, prot, api_key, verify):
        self.url = '{prot}://{host}:{port}/api/v1'.format(prot=prot, host=host, port=port)
        self.session = requests.Session()
        self.session.headers.update({'X-API-Key': api_key})
        self.session.verify = verify

    def _handle_request(self, req):
        if req.status_code in [200, 201, 204]:
            if req.text:
                try:
                    return json.loads(req.text)
                except Exception as e:
                    print(e)  # same as yield
            return dict()
        elif req.status_code == 404:
            error_message = 'Not found'
        else:
            error_message = self._get_request_error_message(data=req)

        raise PowerDNSError(url=req.url,
                            status_code=req.status_code,
                            message=error_message)

    def _get_request_error_message(self, data):
        request_json = data.json()
        if 'error' in request_json:
            request_error = request_json.get('error')
        elif 'errors' in request_json:
            request_error = request_json.get('errors')
        else:
            request_error = 'No error message found'
        return request_error

    def _get_zones_url(self, server):
        return '{url}/servers/{server}/zones'.format(url=self.url, server=server)

    def _get_zone_url(self, server, name):
        return '{url}/{name}'.format(url=self._get_zones_url(server), name=name)

    def get_zone(self, server, name):
        req = self.session.get(url=self._get_zone_url(server, name))
        if req.status_code == 422:  # zone does not exist
            return None
        return self._handle_request(req)

    def get_record(self, server, zone, name):
        # TODO: implementation
        return dict()

    def _get_request_data(self, changetype, server, zone, name, rtype, content=None, disabled=None, ttl=None):
        record_content = list()
        record_content.append(dict(content=content, disabled=disabled))
        record = dict(name=name, type=rtype, changetype=changetype, records=record_content, ttl=ttl)
        rrsets = list()
        rrsets.append(record)
        data = dict(rrsets=rrsets)
        return data

    def create_record(self, server, zone, name, rtype, content, disabled, ttl):
        url = self._get_zone_url(server=server, name=zone)
        data = self._get_request_data(
            changetype='REPLACE',
            server=server,
            zone=zone,
            name=name,
            rtype=rtype,
            content=content if rtype != 'TXT' else '"{}"'.format(content),
            disabled=disabled,
            ttl=ttl
        )
        req = self.session.patch(url=url, json=data)
        return self._handle_request(req)

    def delete_record(self, server, zone, name, rtype):
        url = self._get_zone_url(server=server, name=zone)
        data = self._get_request_data(changetype='DELETE', server=server, zone=zone, name=name, rtype=rtype)
        # module.fail_json(msg=json.dumps(data))
        req = self.session.patch(url=url, json=data)
        return self._handle_request(req)


def ensure(module, pdns_client):
    content = module.params['content']
    disabled = module.params['disabled']
    name = module.params['name']
    rtype = module.params['type']
    ttl = module.params['ttl']
    zone_name = module.params['zone']

    if zone_name not in name:
        name = '{name}.{zone}'.format(name=name, zone=zone_name)
    server = module.params['server']
    state = module.params['state']

    try:
        zone = pdns_client.get_zone(server, zone_name)
    except PowerDNSError as e:
        module.fail_json(
            msg='Could not get zone {name}: HTTP {code}: {err}'.format(name=zone_name, code=e.status_code,
                                                                       err=e.message))

    if not zone:
        module.fail_json(msg='Zone not found: {name}'.format(zone=zone_name))

    # Try to find the record by name and type
    record = next((item for item in zone.get('rrsets') if (item['name'] == name and item['type'] == rtype)), None)

    if state == 'present':
        # Create record if it does not exist
        if not record:
            try:
                pdns_client.create_record(server=server, zone=zone_name, name=name, rtype=rtype, content=content,
                                          ttl=ttl, disabled=disabled)
                return True, pdns_client.get_record(server=server, zone=zone_name, name=name)
            except PowerDNSError as e:
                module.fail_json(
                    msg='Could not create record {name}: HTTP {code}: {err}'.format(name=name, code=e.status_code,
                                                                                    err=e.message))
        # Check if changeable parameters match, else update record.
        if content not in [c.get('content') for c in record["records"]] or record.get('ttl', None) != ttl:
            try:
                pdns_client.create_record(server=server, zone=zone_name, name=name, rtype=rtype, content=content,
                                          ttl=ttl, disabled=disabled)
                return True, pdns_client.get_record(server=server, zone=zone_name, name=name)
            except PowerDNSError as e:
                module.fail_json(
                    msg='Could not update record {name}: HTTP {code}: {err}'.format(name=name, code=e.status_code,
                                                                                    err=e.message))
    elif state == 'absent':
        if record:
            try:
                pdns_client.delete_record(server=server, zone=zone_name, name=name, rtype=rtype)
                return True, None
            except PowerDNSError as e:
                module.fail_json(
                    msg='Could not delete record {name}: HTTP {code}: {err}'.format(name=name, code=e.status_code,
                                                                                    err=e.message))

    return False, record


def main():
    module = AnsibleModule(
        argument_spec=dict(
            content=dict(type='str', required=False),
            disabled=dict(type='bool', default=False),
            name=dict(type='str', required=True),
            server=dict(type='str', default='localhost'),
            state=dict(type='str', default='present', choices=['present', 'absent']),
            ttl=dict(type='int', default=86400),
            type=dict(type='str', required=False, choices=['A', 'AAAA', 'CNAME', 'MX', 'PTR', 'SOA', 'SRV', 'TXT', 'LUA']),
            zone=dict(type='str', required=True),
            pdns_host=dict(type='str', default='127.0.0.1'),
            pdns_port=dict(type='int', default=8081),
            pdns_prot=dict(type='str', default='http', choices=['http', 'https']),
            pdns_api_key=dict(type='str', required=False),
            strict_ssl_checking=dict(type='bool', default=True),
        ),
        supports_check_mode=True,
    )

    pdns_client = PowerDNSClient(host=module.params['pdns_host'],
                                 port=module.params['pdns_port'],
                                 prot=module.params['pdns_prot'],
                                 api_key=module.params['pdns_api_key'],
                                 verify=module.params['strict_ssl_checking'])

    try:
        changed, record = ensure(module, pdns_client)
        module.exit_json(changed=changed, record=record)
    except Exception as e:
        module.fail_json(msg='Error: {0}'.format(str(e)))


# import module snippets
from ansible.module_utils.basic import AnsibleModule, json # noqa

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2018 Procube Co., Ltd.

import docker
import logging
from logging import handlers
import json
import os

# logging.basicConfig(level=os.environ.get('HIVE_LOG_LEVEL', logging.INFO))
DAEMON = None


def volumes(client, logger):
  try:
    for service in client.services.list():
      logger.debug(f'traverse volumes in {service.name}')
      for volume in service.attrs.get('Spec', {}).get('TaskTemplate', {}).get('ContainerSpec', {}).get('Mounts', []):
        if volume.get('Type') == 'volume':
          logger.debug(f'found volume: {json.dumps(volume)}')
          yield {'{#SERVICE_NAME}': service.name, '{#NAME}': volume.get('Source'), '{#PATH}': volume.get('Target')}
  except Exception as e:
    logger.exception(f'fail to get services', e)


def main():
  logger = logging.getLogger('docker-volume-discover')
  logger.setLevel(os.environ.get('HIVE_LOG_LEVEL', logging.INFO))
  syslog_handler = handlers.SysLogHandler(address="/dev/log", facility=handlers.SysLogHandler.LOG_LOCAL1)
  logger.addHandler(syslog_handler)
  try:
    client = docker.from_env()
    print(json.dumps(dict(data=[v for v in volumes(client, logger)])))
  except docker.errors.APIError as e:
    logger.exception(f'fail to initialize docker client: {e}')


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2018 Procube Co., Ltd.

"""
hive inventory: dynamic inventory plugin for hive
"""

from ansible.plugins.inventory import BaseInventoryPlugin
from ansible.errors import AnsibleParserError
import ipaddress

DOCUMENTATION = r'''
  name: hive_inventory
  plugin_type: inventory
  short_description: Returns Inventory for hive docker site
  description:
    - Generate vpcs, firewall, subnets, hosts resources

  options:
    plugin:
      description: token that ensures this is a source file for the 'hive_inventory' plugin.
      required: True
      choices: ['hive_inventory']
    name:
      description:
      - name of hive
      - all hosts can reffernce as hive_name
      - domein name format(ex. pdns.example.com) is recommended
      required: True
    stages:
      description: stage list
      required: True
      suboptions:
        stage:
          description: Name of the stage
          required: true
          choices: ['private', 'staging', 'production']
        provider:
          description: infrastructure provider
          required: true
          choices: ['vagrant', 'aws', 'azure', 'gcp', 'openstack', 'prepared', 'kickstart']
        separate_repository:
          description: whether repository node is separated from swarm nodes
          type: bool
          default: true
        cidr:
          description: cidr of vpc
          required: true
        # inbound_rules:
        #   description: inbound rules
        #   type: list
        #   suboptions:
        #     port:
        #       description: port number
        #       required: true
        #       type: int
        #     src:
        #       description: network addresses to ristrict source ip
        #       type: list
        #       default: ['0.0.0.0/0']
        number_of_hosts:
          description: number of hosts
          default: 4 if separate_repository else 3
          type: int
        subnets:
          description: list of subnet
          type: list
          suboptions:
            cidr:
              description: cidr of subnet
              required: true
            name:
              description: name of subnet
              default: "{name of site}_subnet_{index of item}"
            available_zone:
              description: available zone of subnet
              required: true
        memory_size:
          description: "memory size of hive hosts(only available when provider is 'vagrant')"
          type: int with unit(G,M,K)
          default: VirtualBox default
        cpus:
          description: "number of cpu of hive hosts(only available when provider is 'vagrant')"
          type: int
          default: VirtualBox default
        root_password:
          description: "password of root user (only available when provider is 'preapred')"
        instance_type:
          description: "the instance type of hive hosts(availabe when provider is a IaaS)"
        image_name:
          description: "the source image of the hive host(availabe when provider is a IaaS)"
        region:
          description: "the region where hive hosts are located (availabe when provider is a IaaS)"
        disk_size:
          description: disk size of hive hosts
          type: int with unit(G,M,K)
          default: 20G
        mirrored_disk_size:
          description: disk size of hosts for drbd mirror disk, if not specified then hive does not provision mirrord disk
          type: int (megabyte)
        repository_memory_size:
          description: "memory size of hosts(only available when provider is 'vagrant')"
          type: int with unit(G,M,K)
          default: 512M
        repository_instance_type:
          description: "instance type of host(availabe when provider is a IaaS)"
        repository_disk_size:
          description: disk size of hosts
          type: int with unit(G,M,K)
          default: 40G
'''

EXAMPLES = r'''
plugin: hive_inventory
name: test.example.com
stages:
  private:
    provider: vagrant
    separate_repository: False
    subnets:
    - cidr: 192.168.0.96/29
'''


class InventoryModule(BaseInventoryPlugin):

  NAME = 'hive_inventory'  # used internally by Ansible, it should match the file name but not required

  def __init__(self):
    super(InventoryModule, self).__init__()
    self.sites = []

  def verify_file(self, path):
    ''' return true/false if this is possibly a valid file for this plugin to consume '''
    valid = False
    if super(InventoryModule, self).verify_file(path):
      # base class verifies that file exists and is readable by current user
      if path.endswith(('.yaml', '.yml')):
        valid = True
    return valid

  def parse(self, inventory, loader, path, cache=True):
    super(InventoryModule, self).parse(inventory, loader, path, cache)

    self._read_config_data(path)
    self.inventory.add_group('servers')
    self.inventory.add_group('hives')
    self.inventory.add_group('repository')
    self.inventory.add_group('first_hive')
    self.inventory.add_group('mother')
    self.inventory.add_group('hosts')
    self.inventory.add_child('servers', 'hives')
    self.inventory.add_child('servers', 'repository')
    self.inventory.add_child('hosts', 'servers')
    self.inventory.add_child('hosts', 'mother')
    self.name = self.get_option('name')
    self.inventory.set_variable('all', 'hive_name', self.name)
    stages = self.get_option('stages')
    for stage_name, option in stages.items():
      stage = Stage(self.name, stage_name, option, inventory)
      stage.set_provider()
      stage.add_stage_group()
      stage.set_subnets()
      stage.add_hives()


class Stage:
  def __init__(self, name, stage_name, option, inventory):
    self.stage_name = stage_name
    self.name = name
    self.stage = option
    self.inventory = inventory
    if self.stage_name not in ['private', 'staging', 'production']:
      raise AnsibleParserError(f'stage must be one of "private", "staging", "production", but specified {self.stage_name}')
    self.stage_prefix = 'p-' if self.stage_name == 'private' else 's-' if self.stage_name == 'staging' else ''

  def set_provider(self):
    if 'provider' not in self.stage:
      raise AnsibleParserError('provider must be specified')
    self.provider = self.stage['provider']
    if self.provider not in ['vagrant', 'aws', 'azure', 'gcp', 'openstack', 'prepared', 'kickstart']:
      raise AnsibleParserError(f'provider must be one of "vagrant", "aws", "azure", "gcp", "openstack", "prepared", "kickstart", but specified {self.provider}')
    if self.provider == 'vagrant':
      if 'instance_type' in self.stage:
        raise AnsibleParserError('instance_type cannot be specified when provider is vagrant')
      if 'region' in self.stage:
        raise AnsibleParserError('region cannot be specified when provider is vagrant')
    else:
      if 'memory_size' in self.stage:
        raise AnsibleParserError('memory_size cannot be specified when provider is IaaS')
      if 'repository_memory_size' in self.stage:
        raise AnsibleParserError('repository_memory_size cannot be specified when provider is IaaS')
      if 'dev' in self.stage:
        raise AnsibleParserError('dev cannot be specified when provider is IaaS')
      if 'bridge' in self.stage:
        raise AnsibleParserError('bridge cannot be specified when provider is IaaS')

  def add_stage_group(self):
    self.inventory.add_group(self.stage_name)
    self.inventory.set_variable(self.stage_name, 'hive_provider', self.provider)

  def set_subnets(self):
    self.subnets = []
    if 'cidr' not in self.stage:
      raise AnsibleParserError('cidr must be specified')
    mother_name = f'{self.stage_prefix}mother.{self.name}'
    self.inventory.add_host(mother_name, group='mother')
    self.inventory.add_host(mother_name, group=self.stage_name)
    self.inventory.set_variable(mother_name, 'hive_cidr', self.stage['cidr'])
    if 'region' in self.stage:
      self.inventory.set_variable(mother_name, 'hive_region', self.stage['region'])
    if 'bridge' in self.stage:
      self.inventory.set_variable(mother_name, 'hive_bridge', self.stage['bridge'])
    if 'dev' in self.stage:
      self.inventory.set_variable(mother_name, 'hive_dev', self.stage['dev'])
    if 'nameservers' in self.stage:
      self.inventory.set_variable(mother_name, 'hive_nameservers', self.stage['nameservers'])
    if 'subnets' not in self.stage:
      try:
        net = ipaddress.ip_network(self.stage['cidr'])
      except ValueError as e:
        raise AnsibleParserError(str(e))
      default_subnet = {'cidr': self.stage['cidr'], 'name': self.name + '-default'}
      self.inventory.set_variable(mother_name, 'hive_subnets', [default_subnet.copy()])
      default_subnet['netmask'] = str(net.netmask)
      if 'ip_address_list' in self.stage:
        default_subnet['ip_list'] = (y for y in self.stage.get('ip_address_list'))
      else:
        hosts = net.hosts()
        # first one is route, so skip it
        next(hosts)
        # azure use 2 more address https://docs.microsoft.com/en-us/azure/virtual-network/virtual-networks-faq
        next(hosts)
        next(hosts)
        default_subnet['ip_list'] = map(str, hosts)
      self.subnets.append(default_subnet)
    else:
      var_subnets = []
      for idx, s in enumerate(self.stage['subnets']):
        subnet = s.copy()
        var_subnets.append(s)
        if 'name' not in subnet:
          subnet['name'] = f'{self.stage_prefix}subnet{idx}'
        if 'cidr' not in subnet:
          raise AnsibleParserError('cidr in subnet must be specified')
        try:
          net = ipaddress.ip_network(subnet['cidr'])
        except ValueError as e:
          raise AnsibleParserError(str(e))
        hosts = net.hosts()
        # first one is route, so skip it
        next(hosts)
        # aws use 3 more address https://aws.amazon.com/jp/vpc/faqs/
        next(hosts)
        next(hosts)
        next(hosts)
        subnet['ip_list'] = map(str, hosts)
        subnet['netmask'] = str(net.netmask)
        self.subnets.append(subnet)
      self.inventory.set_variable(mother_name, 'hive_subnets', var_subnets)

  def add_hives(self):
    separate_repository = self.stage.get('separate_repository', True)
    number_of_hosts = self.stage.get('number_of_hosts', 4 if separate_repository else 3)
    if 'ip_address_list' in self.stage:
      number_of_hosts = len(self.stage.get('ip_address_list'))
    for idx in range(number_of_hosts):
      host_name = f'{self.stage_prefix}hive{idx}.{self.name}'
      self.inventory.add_host(host_name, group=self.stage_name)
      if 'root_password' in self.stage:
        self.inventory.set_variable(host_name, 'hive_root_password', self.stage['root_password'])
      if 'internal_cidr' in self.stage:
        self.inventory.set_variable(host_name, 'hive_internal_cidr', self.stage['internal_cidr'])
      if idx == number_of_hosts - 1:
        if not separate_repository:
          self.inventory.add_host(host_name, group='hives')
          if 'memory_size' in self.stage:
            self.inventory.set_variable(host_name, 'hive_memory_size', self.stage['memory_size'])
          if 'cpus' in self.stage:
            self.inventory.set_variable(host_name, 'hive_cpus', self.stage['cpus'])
          if 'disk_size' in self.stage:
            self.inventory.set_variable(host_name, 'hive_disk_size', self.stage['disk_size'])
          if 'instance_type' in self.stage:
            self.inventory.set_variable(host_name, 'hive_instance_type', self.stage['instance_type'])
          if 'mirrored_disk_size' in self.stage:
            self.inventory.set_variable(host_name, 'hive_mirrored_disk_size', self.stage['mirrored_disk_size'])
        self.inventory.add_host(host_name, group='repository')
        if 'repository_memory_size' in self.stage:
          self.inventory.set_variable(host_name, 'hive_memory_size', self.stage['repository_memory_size'])
        if 'repository_cpus' in self.stage:
          self.inventory.set_variable(host_name, 'hive_cpus', self.stage['repository_cpus'])
        if 'repository_disk_size' in self.stage:
          self.inventory.set_variable(host_name, 'hive_disk_size', self.stage['repository_disk_size'])
        if 'repository_instance_type' in self.stage:
          self.inventory.set_variable(host_name, 'hive_instance_type', self.stage['repository_instance_type'])
      else:
        self.inventory.add_host(host_name, group='hives')
        if 'memory_size' in self.stage:
          self.inventory.set_variable(host_name, 'hive_memory_size', self.stage['memory_size'])
        if 'cpus' in self.stage:
          self.inventory.set_variable(host_name, 'hive_cpus', self.stage['cpus'])
        if 'disk_size' in self.stage:
          self.inventory.set_variable(host_name, 'hive_disk_size', self.stage['disk_size'])
        if 'instance_type' in self.stage:
          self.inventory.set_variable(host_name, 'hive_instance_type', self.stage['instance_type'])
        if 'mirrored_disk_size' in self.stage:
          self.inventory.set_variable(host_name, 'hive_mirrored_disk_size', self.stage['mirrored_disk_size'])

      if 'image_name' in self.stage:
        self.inventory.set_variable(host_name, 'hive_image_name', self.stage['image_name'])

      subnet = self.subnets[idx % len(self.subnets)]
      if 'name' in subnet:
        self.inventory.set_variable(host_name, 'hive_subnet', subnet['name'])
      if 'region' in self.stage:
        az_suffix_list = self.stage.get('az_suffix_list', ['-a', '-b', '-c'])
        az_default = self.stage['region'] + az_suffix_list[idx % len(az_suffix_list)]
        if self.provider == 'azure':
          az_default = (idx % 3) + 1
        self.inventory.set_variable(host_name, 'hive_available_zone', subnet.get('available_zone', az_default))
      self.inventory.set_variable(host_name, 'hive_private_ip', next(subnet['ip_list']))
      self.inventory.set_variable(host_name, 'hive_netmask', subnet['netmask'])
    self.inventory.set_variable('hives', 'hive_swarm_master', f'{self.stage_prefix}hive0.{self.name}')
    self.inventory.add_host(f'{self.stage_prefix}hive0.{self.name}', group='first_hive')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2018 Procube Co., Ltd.

"""
hive: support tool to build docker site
"""
import argparse
import configparser
import subprocess
import os
import yaml
from logging import getLogger, StreamHandler, DEBUG, INFO
import traceback
import sys
import re
import signal
import time
import pathlib
import json


def get_python_path():
  for path in os.getenv("PATH").split(os.path.pathsep):
    full_path = os.path.join(path, 'python')
    if os.path.isfile(full_path) and os.access(full_path, os.X_OK):
      return full_path


class Error(Exception):
  pass


def directory_cast(directory):
  if directory.endswith('/'):
    directory = directory[:-1]
  if not os.path.isdir(directory):
    raise Error(f'{directory} is not a directory')
  return os.path.abspath(directory)


def int_cast(int_string):
  try:
    return int(int_string)
  except ValueError:
    raise Error(f'cannot parse as integer {int_string}')


def enum_cast(value, options):
  if value not in options:
    raise Error(f'value {value} not in {options}')
  return value


def bool_cast(s):
  if s.lower() in ["true", "yes", "1"]:
    return True
  if s.lower() in ["false", "no", "0"]:
    return False
  raise Error(f'boolean value {s} not in ["true", "yes", "1", "false", "no", "0"]')


def dict_cast(s):
  return json.loads(s)


class commandHandlerBase:
  def __init__(self, name, description):
    self.name = name
    self.description = description
    self.logger = getLogger('hive')

  def setup_parser(self, subparsers, variables_metainf):
    self.subparser = subparsers.add_parser(
        self.name, description=self.description, help='see `{0} -h`'.format(self.name))
    self.subparser.set_defaults(handler=self)


class initializeEnvironment(commandHandlerBase):
  def __init__(self):
    super().__init__('init', 'initialize hive environment')

  def do(self, context):
    if context.initialize():
      self.logger.info('Success to initialize hive context on {0}'.format(
          f'{context.vars["root_dir"]}/.hive'))
    else:
      self.logger.info('hive context on {0} is already initialized'.format(
          f'{context.vars["root_dir"]}/.hive'))


class setPersistent(commandHandlerBase):
  def __init__(self):
    super().__init__('set', 'set hive variable persistently')

  def do(self, context):
    name = context.args.variable_name
    casted_value = context.get_cast(name)(context.args.value)
    scope = context.set_persistent(name, casted_value)
    context.save_persistent()
    context.logger.info(f'set persistently {casted_value} to {name} on {scope}')

  def setup_parser(self, subparsers, variables_metainf):
    super().setup_parser(subparsers, variables_metainf)
    self.subparser.add_argument('variable_name', help='variable name')
    self.subparser.add_argument('value', help='variable value')


class hiveContext:
  def __init__(self):
    self.init_logger()

  def setup(self, subcommand_handlers):
    self.load_variables_metainf()
    if subcommand_handlers:
      self.args = self.get_parser(subcommand_handlers).parse_args()
    else:
      self.args = argparse.Namespace()
      self.args.root_dir = os.getcwd()
      self.args.verbose = False
    # raise SystemExit when invalid argument or -h option is specified
    self.reset_logger()
    self.vars = {'root_dir': self.args.root_dir,
                 'install_dir': os.path.abspath(os.path.dirname(__file__)),
                 'local_python_path': get_python_path()}
    self.load_persistents()
    self.load_command_line_options()
    self.overwritables = []
    self.load_defaults()
    self.overwritables = []
    self.logger.debug(f'varaiables initialized as {self.vars}')

  def load_variables_metainf(self):
    variables_metainf_path = os.path.dirname(
        __file__) + '/variables_metainf.yml'
    if os.path.exists(variables_metainf_path):
      with open(variables_metainf_path, 'r') as envf:
        self.variables_metainf = yaml.load(envf, Loader=yaml.SafeLoader)
      self.logger.debug('variables meta information {0} is loaded from {1}'.format(
          self.variables_metainf, variables_metainf_path))

  def load_persistents(self):
    # persistent format
    # varname:
    #   global: global value
    #   private: value for private stage
    #   staging: value for staging stage
    #   production: value for production stage
    persistent_values_path = f'{self.vars["root_dir"]}/.hive/persistents.yml'
    if os.path.exists(persistent_values_path):
      with open(persistent_values_path, 'r+') as envf:
        self.persistent_values = yaml.load(envf, Loader=yaml.SafeLoader)
      self.logger.debug('persistent values {0} is loaded from {1}'.format(
          self.persistent_values, persistent_values_path))
    else:
      self.persistent_values = {}
    # update from persistent values
    self.vars.update(dict(
        (k, v['global']) for k, v in self.persistent_values.items() if 'global' in v))
    if 'stage' not in self.vars:
      self.vars['stage'] = self.variables_metainf['stage']['default']
    # update from stage specific value
    self.vars.update(dict((k, v[self.vars['stage']])
                          for k, v in self.persistent_values.items() if self.vars['stage'] in v))

  def save_persistent(self):
    persistent_values_path = f'{self.vars["root_dir"]}/.hive/persistents.yml'
    with open(persistent_values_path, 'w') as envf:
      yaml.dump(self.persistent_values, envf, default_flow_style=False)

  def set_persistent(self, name, value):
    self.initialize()
    if name not in self.variables_metainf:
      raise Error(f'unknown variable {name}')
    metainf = self.variables_metainf[name]
    if ('persistent_scope' in metainf) and metainf['persistent_scope'] == 'none':
      raise Error(f'variable {name} can not be saved persistently')
    persistent_scope = metainf['persistent_scope'] if 'persistent_scope' in metainf else 'global'
    key = self.vars['stage'] if persistent_scope == 'stage' else 'global'
    if name not in self.persistent_values:
      self.persistent_values[name] = {}
    self.persistent_values[name][key] = value
    return key

  def load_defaults(self):
    vars_defaults = [(k, v['default'])
                     for k, v in self.variables_metainf.items() if 'default' in v]
    dirty = True
    while (dirty):
      dirty = False
      for k, v in vars_defaults:
        new_value = v
        try:
          if type(v) == str:
            new_value = v.format(**self.vars)
        except Exception:
          pass
        else:
          if k not in self.vars or (k in self.overwritables and self.vars[k] != new_value):
            self.vars[k] = new_value
            if k not in self.overwritables:
              self.overwritables.append(k)
            dirty = True

  def initialize(self):
    path = f'{self.vars["root_dir"]}/.hive'
    if not os.path.exists(path):
      os.mkdir(path)
      return True
    return False

  def get_cast(self, name):
    if name not in self.variables_metainf:
      raise Error(f'unknown variable {name}')
    metainf = self.variables_metainf[name]
    if 'type' in metainf and metainf['type'] == 'boolean':
      return bool_cast
    if 'type' in metainf and metainf['type'] == 'enum':
      return lambda x: enum_cast(x, metainf['options'])
    if 'type' in metainf and metainf['type'] == 'integer':
      return int_cast
    if 'type' in metainf and metainf['type'] == 'directory':
      return directory_cast
    if 'type' in metainf and metainf['type'] == 'dict':
      return dict_cast
    return str

  @staticmethod
  def add_argument(parser, name, metainf, mutually_exclusive_group_map, help_format_dict):
    long_option = f'--{name.replace("-","_")}'
    kwargs = {'help': metainf['description'].format(**help_format_dict)}
    if 'type' in metainf and metainf['type'] == 'boolean':
      kwargs['action'] = 'store_true'
    if 'type' in metainf and metainf['type'] == 'enum':
      kwargs['choices'] = metainf['options']
    if 'type' in metainf and metainf['type'] == 'integer':
      kwargs['type'] = int_cast
    if 'type' in metainf and metainf['type'] == 'directory':
      kwargs['type'] = directory_cast
    target = parser
    if 'command_line_mutually_exclusive_group' in metainf:
      group_name = metainf['command_line_mutually_exclusive_group']
      if group_name not in mutually_exclusive_group_map:
        mutually_exclusive_group_map[group_name] = parser.add_mutually_exclusive_group()
      target = mutually_exclusive_group_map[group_name]
    target.add_argument(metainf['command_line_option'], long_option, **kwargs)

  def get_parser(self, subcommand_handlers):
    parser = argparse.ArgumentParser(
        description='support tool to build docker site')
    mutually_exclusive_group_map = {}
    for metainf in [(k, v) for k, v in self.variables_metainf.items()
                    if 'command_line_option' in v and (('command_line_option_level' not in v) or v['command_line_option_level'] == 'global')]:
      hiveContext.add_argument(
          parser, metainf[0], metainf[1], mutually_exclusive_group_map, {})
    parser.set_defaults(root_dir=os.getcwd())
    subparsers = parser.add_subparsers(title='subcommands')
    for handler in subcommand_handlers:
      handler.setup_parser(subparsers, self.variables_metainf)
    return parser

  def load_command_line_options(self):
    self.vars.update(dict((k, getattr(self.args, k)) for k, v in self.variables_metainf.items()
                          if k != 'root_dir' and 'command_line_option' in v and hasattr(self.args, k) and getattr(self.args, k)))

  def do_subcommand(self):
    self.args.handler.do(self)

  def init_logger(self):
    self.logger = getLogger('hive')
    self.standard_handler = StreamHandler()
    self.logger.addHandler(self.standard_handler)
    self.logger.propagate = False
    self.standard_handler.setLevel(INFO)
    self.logger.setLevel(INFO)

  def reset_logger(self):
    if self.args.verbose:
      self.standard_handler.setLevel(DEBUG)
      self.logger.setLevel(DEBUG)


def run_and_check_ansible_playbook(args):
  proc = subprocess.run(args)
  if proc.returncode != 0:
    raise Error(f'ansible-playbook command failed: exit code = {proc.returncode}')


HOST_DEF_PATTERN = re.compile('^Host *(.*)$')


class ansbileCommandBase(commandHandlerBase):
  def get_playbook_path(self, context):
    return f'{context.vars["playbooks_dir"]}/{self.name}.yml'

  def collect_ansible_vars(self, vars, variables_metainf):
    # override when phase sapecific ansible variale
    ansible_vars = dict((v['name_in_ansible'], vars[k])
                        for k, v in variables_metainf.items() if k in vars and 'name_in_ansible' in v)
    # ansible_vars['hive_temp_dir'] = f'{vars["root_dir"]}/{vars["stage"]}/.wk_{self.name}'
    return ansible_vars

  def collect_ansible_cfg_vars(self, vars, variables_metainf):
    # override when phase sapecific ansible variale
    ansible_cfg_vars = [dict(name=v['name_in_ansible_cfg'], value=vars[k], section=v['section_in_ansible_cfg'] if 'section_in_ansible_cfg' in v else 'defaults')
                        for k, v in variables_metainf.items() if k in vars and 'name_in_ansible_cfg' in v]
    return ansible_cfg_vars

  def collect_environment_vars(self, vars, variables_metainf, ansible_cfg_file_path):
    # override when phase sapecific ansible variale
    environment_vars = dict((v['name_in_environment'], vars[k])
                            for k, v in variables_metainf.items() if k in vars and 'name_in_environment' in v)
    environment_vars['ANSIBLE_CONFIG'] = ansible_cfg_file_path
    return environment_vars

  def build_context(self, context):
    context.load_defaults()
    context_dir = context.vars["context_dir"]
    temp_dir = context.vars["temp_dir"]
    os.makedirs(context_dir, exist_ok=True)
    os.makedirs(temp_dir, exist_ok=True)
    ansible_vars = self.collect_ansible_vars(
        context.vars, context.variables_metainf)
    self.vars_file_path = f'{temp_dir}/vars.yml'
    context.logger.debug(f'ansible extra variables: {ansible_vars}')
    with open(self.vars_file_path, 'w') as varsf:
      yaml.dump(ansible_vars, varsf, default_flow_style=False)
      ansible_cfg_vars = self.collect_ansible_cfg_vars(
          context.vars, context.variables_metainf)
      context.logger.debug(f'ansible cfg setting {ansible_cfg_vars}')
      # default interpolation cause error by % char, which is used by ssh_args
      config = configparser.ConfigParser(
          interpolation=configparser.ExtendedInterpolation())
      for var in ansible_cfg_vars:
        if not config.has_section(var['section']):
          config.add_section(var['section'])
        config.set(var['section'], var['name'], str(var['value']))
      ansible_cfg_file_path = f'{temp_dir}/ansible.cfg'
      with open(ansible_cfg_file_path, 'w') as cfg_varsf:
        config.write(cfg_varsf)
      environment_vars = self.collect_environment_vars(
          context.vars, context.variables_metainf, ansible_cfg_file_path)
      for k, v in environment_vars.items():
        os.environ[k] = v
      context.logger.debug(f'environment variables: {os.environ}')

  def setup_parser(self, subparsers, variables_metainf):
    super().setup_parser(subparsers, variables_metainf)
    mutually_exclusive_group_map = {}
    for variable in [(k, v) for k, v in variables_metainf.items() if 'command_line_option' in v and
                     ('command_line_option_level' in v and v['command_line_option_level'] == 'phase') and
                     self.name in v['command_line_option_available']]:
      hiveContext.add_argument(self.subparser, variable[0], variable[1], mutually_exclusive_group_map, {
                               'subject_name': self.subject_name})


class phaseBase(ansbileCommandBase):
  def do(self, context):
    context.vars['phase'] = self.name
    self.build_context(context)
    self.do_one(context)

  def get_limit_targets(self, context):
    return context.vars["stage"]

  def do_one(self, context):
    context.logger.info(f'=== PHASE {self.name} START at {time.strftime("%Y-%m-%d %H:%M:%S %z")} ===')
    args = ['ansible-playbook', '--limit', self.get_limit_targets(context)]
    if 'verbose' in context.vars and context.vars['verbose']:
      args.append('-vvv')
    if 'check_mode' in context.vars and context.vars['check_mode']:
      args.append('-C')
    args += ['--extra-vars', f'@{self.vars_file_path}']
    args.append(self.get_playbook_path(context))
    context.logger.debug(f'commnad={args}')
    run_and_check_ansible_playbook(args)
    context.set_persistent('start_phase', self.name)
    if not context.vars.get('destroy'):
      next_idx = PHASE_LIST.index(self) + 1
      if next_idx < len(PHASE_LIST):
        context.set_persistent('start_phase', PHASE_NAME_LIST[next_idx])
    context.save_persistent()
    context.logger.info(f'=== PHASE {self.name} END at {time.strftime("%Y-%m-%d %H:%M:%S %z")} ===')


class buildInfra(phaseBase):
  def __init__(self):
    super().__init__('build-infra', 'build infrastructure, setup networks, global ip, firewall')
    self.subject_name = 'vpc/subnet/host'


class setupHosts(phaseBase):
  def __init__(self):
    super().__init__('setup-hosts',
                     'setup hosts, install software, configure services, configure cluster')
    self.subject_name = 'host'

  def get_limit_targets(self, context):
    if 'limit_target' in context.vars:
      return ':'.join(context.vars['limit_target'].split(',')) + ':&' + context.vars["stage"]
    return context.vars["stage"]


class phaseWithDockerSocket(phaseBase):

  def do_one(self, context):
    ssh_config_path = context.vars['context_dir'] + '/ssh_config'
    grep_proc = subprocess.run(['grep', '^Host ', ssh_config_path], stdout=subprocess.PIPE)
    if grep_proc.returncode == 1:
      raise Error(f'no Host entry in {ssh_config_path}')
    if grep_proc.returncode != 0:
      raise Error(f'fail to read ssh_config {ssh_config_path} error, you may never done build-infra: {grep_proc.stderr}')
    ssh_tunnel_procs = []
    for host_name in map(lambda x: x.decode(encoding='utf-8').split(' ')[1], grep_proc.stdout.splitlines()):
      socket_path = f'{context.vars["temp_dir"]}/docker.sock@{host_name}'
      if pathlib.Path(socket_path).is_socket():
        raise Error(f'fail to create socket {socket_path}, another hive process may doing build-images/deploy-services' +
                    ' or the file has been left because previus hive process aborted suddenly')
      args = ['ssh', '-N', '-F', ssh_config_path, '-L', socket_path + ':/var/run/docker.sock', host_name]
      ssh_tunnel_procs.append((socket_path, subprocess.Popen(args)))
    try:
      super().do_one(context)
    finally:
      for socket_path, ssh_tunnel_proc in ssh_tunnel_procs:
        ssh_tunnel_proc.send_signal(signal.SIGTERM)
        while ssh_tunnel_proc.poll() is None:
          try:
            ssh_tunnel_proc.wait(timeout=1)
          except subprocess.TimeoutExpired:
            pass
        os.remove(socket_path)


class buildImages(phaseWithDockerSocket):
  def __init__(self):
    super().__init__('build-images', 'build container images')
    self.subject_name = 'container'

  def get_limit_targets(self, context):
    if 'limit_target' in context.vars:
      return ':'.join(map(lambda x: 'image_' + x, context.vars['limit_target'].split(','))) + ':repository:&' + context.vars["stage"]
    return context.vars["stage"]


class buildVolumes(phaseBase):
  def __init__(self):
    super().__init__('build-volumes', 'build volumes on hives')
    self.subject_name = 'volume'


class buildNetworks(phaseBase):
  def __init__(self):
    super().__init__('build-networks', 'build networks for swarm')
    self.subject_name = 'ingress network'


class deployServices(phaseBase):
  def __init__(self):
    super().__init__('deploy-services', 'deploy services')
    self.subject_name = 'service'


class initializeServices(phaseWithDockerSocket):
  def __init__(self):
    super().__init__('initialize-services', 'initialize services')
    self.subject_name = 'service'

  def get_playbook_path(self, context):
    root_dir_path = f'{context.vars["root_dir"]}/{self.name}.yml'
    if os.path.exists(root_dir_path):
      context.logger.warning(
          f'HIVE WARINIG: {self.name}.yml at root directory({context.vars["root_dir"]}) is deprecated. Use initialize_roles property of service definition')
      return root_dir_path
    return super().get_playbook_path(context)

  def get_limit_targets(self, context):
    if 'limit_target' in context.vars:
      return ':'.join(context.vars['limit_target'].split(',')) + ':&' + context.vars["stage"]
    return context.vars["stage"]


PHASE_LIST = [buildInfra(), setupHosts(), buildImages(),
              buildVolumes(), buildNetworks(), deployServices(), initializeServices()]
PHASE_NAME_LIST = list(map(lambda x: x.name, PHASE_LIST))


class allPhase(phaseBase):
  def __init__(self):
    super().__init__('all', 'do all phase')
    self.subject_name = 'all'

  def do(self, context):
    start_phase_idx = PHASE_NAME_LIST.index(context.vars['start_phase'])
    last_idx = len(PHASE_LIST)
    for idx in range(start_phase_idx, last_idx):
      PHASE_LIST[idx].build_context(context)
      PHASE_LIST[idx].do_one(context)


class inventoryList(ansbileCommandBase):
  def __init__(self):
    super().__init__('inventory', 'list ansible inventory')

  def do(self, context):
    self.build_context(context)
    args = ['ansible-inventory', '--playbook-dir=' +
            context.vars['playbooks_dir'], '--list']
    subprocess.run(args)


class execSsh(ansbileCommandBase):
  def __init__(self):
    super().__init__('ssh', 'ssh to hive server')
    self.subject_name = 'host'

  def do(self, context):
    self.build_context(context)
    ssh_config_path = context.vars['context_dir'] + '/ssh_config'
    grep_proc = subprocess.run(['grep', '^Host ', ssh_config_path], stdout=subprocess.PIPE)
    if grep_proc.returncode == 1:
      raise Error(f'no Host entry in {ssh_config_path}')
    if grep_proc.returncode != 0:
      raise Error(f'fail to read ssh_config {ssh_config_path} error, you may never done build-infra: {grep_proc.stderr}')
    hosts = list(map(lambda x: x.decode(encoding='utf-8').split(' ')[1], grep_proc.stdout.splitlines()))
    ssh_host = context.vars.get('ssh_host')
    if ssh_host is None:
      ssh_host = hosts[len(hosts) - 1]
    elif ssh_host not in hosts:
      raise Error(f'host {ssh_host} is not found in {ssh_config_path}')
    args = ['/usr/bin/ssh', '-F', ssh_config_path]
    if context.vars['foward_zabbix']:
      args += ['-L', f'localhost:{context.vars["foward_zabbix_port"]}:{ssh_host}:10052']
    if context.vars.get('port_forwarding'):
      args += ['-L', context.vars['port_forwarding']]
    args += [ssh_host]
    context.logger.debug(f'commnad={args}')
    try:
      ssh_proc = subprocess.Popen(args)
      ssh_proc.wait()
    except KeyboardInterrupt:
      ssh_proc.send_signal(signal.SIGTERM)
      ssh_proc.wait()


SUBCOMMANDS = PHASE_LIST + [allPhase(), inventoryList(), initializeEnvironment(), setPersistent(), execSsh()]


def get_parser():
  context = hiveContext()
  context.load_variables_metainf()
  return context.get_parser(SUBCOMMANDS)


def main():
  context = hiveContext()
  try:
    context.setup(SUBCOMMANDS)
    context.do_subcommand()
  except KeyboardInterrupt:
    context.logger.error('\nABORTED: keyboard interruption')
  except Error as e:
    context.logger.error(f'HIVE ERROR at {time.strftime("%Y-%m-%d %H:%M:%S %z")}: {str(e)}')
    sys.exit(1)
  except Exception:
    print("Exception in hive command:")
    print("-" * 40)
    traceback.print_exc(file=sys.stdout)
    print("-" * 40)
    sys.exit(1)


if __name__ == '__main__':
  main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

DOCUMENTATION = '''
---
module: powerdns_zone
short_description: Manage PowerDNS zones
description:
- Create, update or delete a PowerDNS zone using API
options:
  kind:
    description:
    - Zone kind
    required: False
    default: master
    choices: ['native', 'master', 'slave']
  name:
    description:
    - Zone name
    required: true
  nameservers:
    description:
    - List of nameservers
    required: False
    default: None
  pdns_host:
    description:
    - Name or ip address of PowerDNS host
    required: false
    default: 127.0.0.1
  pdns_port:
    description:
    - Port used by PowerDNS API
    required: false
    default: 8081
  pdns_prot:
    description:
    - Protocol used to connect to PowerDNS API
    required: false
    default: http
    choices: ['http', 'https']
  pdns_api_key:
    description:
    - API Key to authenticate through PowerDNS API
  strict_ssl_checking:
    description:
    - Disables strict certificate checking
    default: false
author: "Thomas Krahn (@nosmoht)"
'''

EXAMPLES = '''
# Ensure a zone is present
- powerdns_zone:
    name: zone01.internal.example.com
    kind: master
    nameservers:
    - ns-01.internal.example.com
    - ns-02.internal.example.com
    state: present
    pdns_host: powerdns.example.cm
    pdns_port: 8080
    pdns_api_key: topsecret

# Ensure a zone is absent
- powerdns_zone:
    name: old-zone.internal.example.com
    state: absent
    pdns_host: powerdns.example.cm
    pdns_port: 8080
    pdns_api_key: topsecret
'''

import requests # noqa


class PowerDNSError(Exception):
    def __init__(self, url, status_code, message):
        self.url = url
        self.status_code = status_code
        self.message = message
        super(PowerDNSError, self).__init__()


class PowerDNSClient:
    def __init__(self, module, host, port, prot, api_key, verify):
        self.module = module
        self.url = '{prot}://{host}:{port}/api/v1'.format(prot=prot, host=host, port=port)
        self.session = requests.Session()
        self.session.headers.update({'X-API-Key': api_key})
        self.session.verify = verify

    def _handle_request(self, req):
        if req.status_code in [200, 201, 204]:
            if req.text:
                try:
                    return json.loads(req.text)
                except Exception as e:
                    print(e)  # same as yield
            return dict()
        elif req.status_code == 404:
            error_message = 'Not found'
        else:
            error_message = self._get_request_error_message(data=req)

        raise PowerDNSError(url=req.url,
                            status_code=req.status_code,
                            message=error_message)

    def _get_request_error_message(self, data):
        request_json = data.json()
        if 'error' in request_json:
            request_error = request_json.get('error')
        elif 'errors' in request_json:
            request_error = request_json.get('errors')
        else:
            request_error = 'DONT KNOW'
        return request_error

    def _get_zones_url(self, server):
        return '{url}/servers/{server}/zones'.format(url=self.url, server=server)

    def _get_zone_url(self, server, name):
        return '{url}/{name}'.format(url=self._get_zones_url(server), name=name)

    def get_zone(self, server, name):
        self.module.debug('HTTP GET {url}'.format(url=self._get_zone_url(server, name)))
        req = self.session.get(url=self._get_zone_url(server, name))
        if req.status_code == 422 or req.status_code == 404:  # zone does not exist
            return None
        return self._handle_request(req)

    def create_zone(self, server, data):
        req = self.session.post(url=self._get_zones_url(server), json=data)
        return self._handle_request(req)

    def delete_zone(self, server, name):
        req = self.session.delete(url=self._get_zone_url(server, name))
        return self._handle_request(req)

    def update_zone(self, server, zone):
        req = self.session.patch(url=self._get_zone_url(server=server, name=zone.get('name')), data=zone)
        return self._handle_request(req)


def diff(list1, list2):
    c = set(list1).union(set(list2))
    d = set(list1).intersection(set(list2))
    return list(c - d)


def ensure(module, pdns_client):
    kind = module.params['kind']
    masters = module.params['masters']
    name = module.params['name']
    nameservers = module.params['nameservers']
    server = module.params['server']
    state = module.params['state']
    try:
        module.debug('get_zone({server},{name})')
        zone = pdns_client.get_zone(server, name)
    except PowerDNSError as e:
        module.fail_json(
            msg='Could not get zone {name}: HTTP {code}: {err}'.format(name=name, code=e.status_code, err=e.message))

    if not zone:
        if state == 'present':
            try:
                zone = dict(name=name, kind=kind, nameservers=nameservers, masters=masters)
                if module.check_mode:
                    module.exit_json(changed=True, zone=zone)
                module.debug('create_zone({server},{name})')
                pdns_client.create_zone(server, zone)
                module.debug('get_zone({server},{name})')
                return True, pdns_client.get_zone(server, name)
            except PowerDNSError as e:
                module.fail_json(
                    msg='Could not create zone {name}: HTTP {code}: {err}'.format(name=name, code=e.status_code,
                                                                                  err=e.message))
    else:
        if state == 'absent':
            try:
                if module.check_mode:
                    module.exit_json(changed=True, zone=zone)
                pdns_client.delete_zone(server, name)  # zone.get('id'))
                return True, None
            except PowerDNSError as e:
                module.fail_json(
                    msg='Could not delete zone {name}: HTTP {code}: {err}'.format(name=name, code=e.status_code,
                                                                                  err=e.message))
                # Compare nameservers
                #        ns_diff = diff(nameservers if nameservers else list(), zone.get('nameservers', list()))
                #        if ns_diff:
                #            try:
                #                if module.check_mode:
                #                    module.exit_json(changed=True, zone=zone)
                #                pdns_client.update_zone(server, zone)
                #                return True, pdns_client.get_zone(name)
                #            except PowerDNSError as e:
                #                module.fail_json(
                #                    msg='Could not update zone {name}: HTTP {code}: {err}'.format(name=name, code=e.status_code,
                #                                                                                  err=e.message))
    return False, zone


def main():
    module = AnsibleModule(
        argument_spec=dict(
            kind=dict(type='str', required=False, default='master', choices=['native', 'master', 'slave']),
            masters=dict(type='list', required=False),
            name=dict(type='str', required=True),
            nameservers=dict(type='list', required=False),
            server=dict(type='str', required=False, default='localhost'),
            state=dict(type='str', default='present', choices=['present', 'absent']),
            pdns_host=dict(type='str', default='127.0.0.1'),
            pdns_port=dict(type='int', default=8081),
            pdns_prot=dict(type='str', default='http', choices=['http', 'https']),
            pdns_api_key=dict(type='str', required=False),
            strict_ssl_checking=dict(type='bool', default=True),
        ),
        supports_check_mode=True,
    )
    pdns_client = PowerDNSClient(module, host=module.params['pdns_host'],
                                 port=module.params['pdns_port'],
                                 prot=module.params['pdns_prot'],
                                 api_key=module.params['pdns_api_key'],
                                 verify=module.params['strict_ssl_checking'])

    try:
        changed, zone = ensure(module, pdns_client)
        module.exit_json(changed=changed, zone=zone)
    except Exception as e:
        module.fail_json(msg='Error: {0}'.format(str(e)))


# import module snippets
from ansible.module_utils.basic import AnsibleModule, json # noqa

if __name__ == '__main__':
    main()
